# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import functools
from typing import (
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Type,
    Union,
)

from ....core import FUSE_CHUNK_TYPE, CHUNK_TYPE, TILEABLE_TYPE
from ....typing import EntityType, TileableType, ChunkType
from ....utils import copy_tileables, build_fetch
from ...entity.tileables import handler
from ...mode import enter_mode
from ..entity import TileableGraph, ChunkGraph
from .base import AbstractGraphBuilder


tile_gen_type = Generator[List[ChunkType], List[ChunkType], List[TileableType]]
DEFAULT_UPDATED_PROGRESS = 0.4


@dataclasses.dataclass
class _TileableHandler:
    tileable: TileableType
    handler: tile_gen_type
    last_need_processes: List[EntityType] = None


@dataclasses.dataclass
class _TileableTileInfo:
    curr_iter: int
    # incremental progress for this iteration
    tile_progress: float
    # newly generated chunks by a tileable in this iteration
    generated_chunks: List[ChunkType] = dataclasses.field(default_factory=list)


class TileContext(Dict[TileableType, TileableType]):
    _tileables = Set[TileableType]
    _tileable_to_progress: Dict[TileableType, float]
    _tileable_to_tile_infos: Dict[TileableType, List[_TileableTileInfo]]

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._tileables = None
        self._tileable_to_progress = dict()
        self._tileable_to_tile_infos = dict()

    def set_tileables(self, tileables: Set[TileableType]):
        self._tileables = tileables

    def __setitem__(self, key, value):
        self._tileable_to_progress.pop(key, None)
        return super().__setitem__(key, value)

    def set_progress(self, tileable: TileableType, progress: float):
        assert 0.0 <= progress <= 1.0
        last_progress = self._tileable_to_progress.get(tileable, 0.0)
        self._tileable_to_progress[tileable] = max(progress, last_progress)

    def get_progress(self, tileable: TileableType) -> float:
        if tileable in self:
            return 1.0
        else:
            return self._tileable_to_progress.get(tileable, 0.0)

    def get_all_progress(self) -> float:
        return sum(self.get_progress(t) for t in self._tileables) / len(self._tileables)

    def record_tileable_tile_info(
        self, tileable: TileableType, curr_iter: int, generated_chunks: List[ChunkType]
    ):
        if tileable not in self._tileable_to_tile_infos:
            self._tileable_to_tile_infos[tileable] = []
        prev_progress = sum(
            info.tile_progress for info in self._tileable_to_tile_infos[tileable]
        )
        curr_progress = self.get_progress(tileable)
        infos = self._tileable_to_tile_infos[tileable]
        infos.append(
            _TileableTileInfo(
                curr_iter=curr_iter,
                tile_progress=curr_progress - prev_progress,
                generated_chunks=generated_chunks,
            )
        )

    def get_tileable_tile_infos(self) -> Dict[TileableType, List[_TileableTileInfo]]:
        return {t: self._tileable_to_tile_infos.get(t, list()) for t in self._tileables}


@dataclasses.dataclass
class TileStatus:
    entities: List[EntityType] = None
    progress: float = None


class Tiler:
    _cur_iter: int
    _cur_chunk_graph: Optional[ChunkGraph]
    _tileable_handlers: Iterable[_TileableHandler]

    def __init__(
        self,
        tileable_graph: TileableGraph,
        tile_context: TileContext,
        processed_chunks: Set[ChunkType],
        chunk_to_fetch: Dict[ChunkType, ChunkType],
        add_nodes: Callable,
    ):
        self._tileable_graph = tileable_graph
        self._tile_context = tile_context
        self._processed_chunks = processed_chunks
        self._chunk_to_fetch = chunk_to_fetch
        self._add_nodes = self._wrap_add_nodes(add_nodes)
        self._curr_iter = 0
        self._cur_chunk_graph = None
        self._tileable_handlers = (
            _TileableHandler(tileable, self._tile_handler(tileable))
            for tileable in tileable_graph.topological_iter()
        )

    def _wrap_add_nodes(self, add_nodes: Callable):
        @functools.wraps(add_nodes)
        def inner(
            chunk_graph: ChunkGraph,
            chunks: List[ChunkType],
            visited: Set[ChunkType],
            tileable: TileableType,
        ):
            prev_chunks = set(chunk_graph)
            add_nodes(chunk_graph, chunks, visited)
            new_chunks = set(chunk_graph)
            self._tile_context.record_tileable_tile_info(
                tileable, self._curr_iter, list(new_chunks - prev_chunks)
            )

        return inner

    @staticmethod
    def _get_data(entity: EntityType):
        return entity.data if hasattr(entity, "data") else entity

    def _tile_handler(self, tileable: TileableType) -> tile_gen_type:
        from ....core.operand import Fetch

        tileable = self._get_data(tileable)

        if isinstance(tileable.op, Fetch) and not tileable.is_coarse():
            return [tileable]

        assert tileable.is_coarse()

        # copy tileable
        tiled_tileables = copy_tileables(
            tileable.op.outputs,
            inputs=[self._tile_context[inp] for inp in tileable.inputs],
            copy_key=True,
            copy_id=False,
        )
        tiled_tileables = [self._get_data(t) for t in tiled_tileables]
        # start to tile
        tiled_tileables = yield from handler.tile(tiled_tileables)
        return tiled_tileables

    def _gen_tileable_handlers(self, next_tileable_handlers: List[_TileableHandler]):
        for tile_handler in self._tileable_handlers:
            tileable, handler = tile_handler.tileable, tile_handler.handler
            if tileable in self._tile_context:
                continue
            if any(
                inp not in self._tile_context
                for inp in self._tileable_graph.predecessors(tileable)
            ):
                # predecessors not finished yet
                next_tileable_handlers.append(_TileableHandler(tileable, handler))
                continue

            yield _TileableHandler(tileable, handler)

    def _tile(
        self,
        chunk_graph: ChunkGraph,
        tileable: TileableType,
        tile_handler: tile_gen_type,
        next_tileable_handlers: List[_TileableHandler],
        to_update_tileables: List[TileableType],
        visited: Set[EntityType],
    ):
        try:
            need_process = next(tile_handler)

            if isinstance(need_process, TileStatus):
                # process tile that returns progress
                self._tile_context.set_progress(tileable, need_process.progress)
                need_process = need_process.entities
            else:
                # if progress not specified, we just update 0.4 * rest progress
                progress = self._tile_context.get_progress(tileable)
                new_progress = progress + (1.0 - progress) * DEFAULT_UPDATED_PROGRESS
                self._tile_context.set_progress(tileable, new_progress)

            chunks = []
            if need_process is not None:
                for t in need_process:
                    if isinstance(t, CHUNK_TYPE):
                        chunks.append(self._get_data(t))
                    elif isinstance(t, TILEABLE_TYPE):
                        to_update_tileables.append(self._get_data(t))
            # not finished yet
            self._add_nodes(chunk_graph, chunks.copy(), visited, tileable)
            next_tileable_handlers.append(
                _TileableHandler(tileable, tile_handler, need_process)
            )
            # add intermediate chunks into result chunks
            # to prevent them being pruned
            chunk_graph.result_chunks.extend(c for c in chunks if c in chunk_graph)
        except StopIteration as e:
            # tile done
            tiled_tileables = e.value
            for out, tiled_tileable in zip(tileable.op.outputs, tiled_tileables):
                out = self._get_data(out)
                tiled_tileable = self._get_data(tiled_tileable)

                chunks = tiled_tileable.chunks
                if chunks is None:  # pragma: no cover
                    raise ValueError(f"tileable({out}) is still coarse after tile")
                chunks = [self._get_data(c) for c in chunks]
                self._tile_context[out] = tiled_tileable
                self._add_nodes(chunk_graph, chunks, visited, tileable)

    def _gen_result_chunks(
        self,
        chunk_graph: ChunkGraph,
        next_tileable_handlers: List[_TileableHandler],
    ):
        result_chunks = chunk_graph.result_chunks
        tileable_graph = self._tileable_graph
        result_chunk_set = set(result_chunks)

        def _add_result_chunk(c):
            if c not in result_chunk_set:
                result_chunks.append(c)
                result_chunk_set.add(c)

        if next_tileable_handlers:
            for tileable_handler in next_tileable_handlers:
                tileable = tileable_handler.tileable
                # tileable that tile not completed, scan their inputs
                for inp_tileable in tileable_graph.iter_predecessors(tileable):
                    if (
                        tileable_handler.last_need_processes is None
                        or tileable_graph.count_successors(inp_tileable) > 1
                    ):
                        # if nothing yielded inside its tile,
                        # or the input has more than 1 successors,
                        # make sure their chunks in result,
                        # so that they will not be executed repeatedly
                        if inp_tileable in self._tile_context:
                            for chunk in self._tile_context[inp_tileable].chunks:
                                chunk = self._get_data(chunk)
                                if chunk in chunk_graph:
                                    _add_result_chunk(chunk)
        for tileable in tileable_graph.result_tileables:
            if tileable in self._tile_context:
                for chunk in self._tile_context[tileable].chunks:
                    chunk = self._get_data(chunk)
                    if chunk in chunk_graph:
                        _add_result_chunk(chunk)
                    if (
                        chunk in self._chunk_to_fetch
                        and self._chunk_to_fetch[chunk] in chunk_graph
                    ):
                        _add_result_chunk(self._chunk_to_fetch[chunk])

    def _iter(self):
        chunk_graph = self._cur_chunk_graph

        to_update_tileables = []
        visited = set()

        if chunk_graph is not None:
            # last tiled chunks, add them to processed
            # so that fetch chunk can be generated
            processed_chunks = [
                c.chunk if isinstance(c, FUSE_CHUNK_TYPE) else c
                for c in chunk_graph.result_chunks
            ]
            self._processed_chunks.update(processed_chunks)

        result_chunks = []
        chunk_graph = self._cur_chunk_graph = ChunkGraph(result_chunks)

        next_tileable_handlers = []
        # tile
        for tile_handler in self._gen_tileable_handlers(next_tileable_handlers):
            self._tile(
                chunk_graph,
                tile_handler.tileable,
                tile_handler.handler,
                next_tileable_handlers,
                to_update_tileables,
                visited,
            )
        self._tileable_handlers = next_tileable_handlers
        # gen result chunks
        self._gen_result_chunks(chunk_graph, next_tileable_handlers)
        # prune unused chunks
        prune_chunk_graph(chunk_graph)

        self._curr_iter += 1

        return to_update_tileables

    def __iter__(self):
        while self._tileable_handlers:
            to_update_tileables = self._iter()
            yield self._cur_chunk_graph
            for t in to_update_tileables:
                t.refresh_params()


def prune_chunk_graph(chunk_graph: ChunkGraph):
    from ....core.operand import Fetch, VirtualOperand, ShuffleProxy

    result_set = set(chunk_graph.result_chunks)
    stack = list(chunk_graph.result_chunks)
    used = set()
    while stack:
        n = stack.pop()
        if n in used:
            continue
        used.add(n)
        stack.extend(chunk_graph.predecessors(n))
        if isinstance(n.op, ShuffleProxy):
            stack.extend(
                succ for succ in chunk_graph.iter_successors(n) if succ not in used
            )

    unused = {n for n in chunk_graph if n not in used}
    for n in unused:
        # for pruned chunks, we assume we will use them later,
        # so we add the inputs of them into result chunks,
        # to prevent from duplicated submission
        for inp in chunk_graph.iter_predecessors(n):
            if (
                inp in used
                and inp not in result_set
                and not isinstance(inp.op, (Fetch, VirtualOperand))
            ):
                chunk_graph.result_chunks.append(inp)
                result_set.add(inp)
        # prune chunk
        chunk_graph.remove_node(n)


class ChunkGraphBuilder(AbstractGraphBuilder):
    _graph: TileableGraph

    def __init__(
        self,
        graph: TileableGraph,
        fuse_enabled: bool = True,
        tile_context: TileContext = None,
        tiler_cls: Union[Type[Tiler], Callable] = None,
    ):
        super().__init__(graph)
        self.fuse_enabled = fuse_enabled
        self.tile_context = TileContext() if tile_context is None else tile_context
        self.tile_context.set_tileables(set(graph))

        self._processed_chunks: Set[ChunkType] = set()
        self._chunk_to_fetch: Dict[ChunkType, ChunkType] = dict()

        tiler_cls = Tiler if tiler_cls is None else tiler_cls
        self.tiler = tiler_cls(
            self._graph,
            self.tile_context,
            self._processed_chunks,
            self._chunk_to_fetch,
            self._add_nodes,
        )

    def _process_node(self, entity: EntityType):
        if entity in self._processed_chunks:
            if entity not in self._chunk_to_fetch:
                # gen fetch
                fetch_chunk = build_fetch(entity).data
                self._chunk_to_fetch[entity] = fetch_chunk
            return self._chunk_to_fetch[entity]
        return entity

    def _select_inputs(self, inputs: List[ChunkType]):
        new_inputs = []
        for inp in inputs:
            if inp in self._processed_chunks:
                # gen fetch
                if inp not in self._chunk_to_fetch:
                    fetch_chunk = build_fetch(inp).data
                    self._chunk_to_fetch[inp] = fetch_chunk
                new_inputs.append(self._chunk_to_fetch[inp])
            else:
                new_inputs.append(inp)
        return new_inputs

    def _if_add_node(self, node: EntityType, visited: Set):
        return node not in visited and node not in self._processed_chunks

    def _build(self) -> Iterable[Union[TileableGraph, ChunkGraph]]:
        tile_iterator = iter(self.tiler)
        while True:
            try:
                with enter_mode(build=True, kernel=True):
                    graph = next(tile_iterator)
                yield graph
            except StopIteration:
                break

    def build(self) -> Generator[Union[TileableGraph, ChunkGraph], None, None]:
        yield from self._build()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections.abc import Sequence
from typing import List

import numpy as np
from scipy.sparse.base import spmatrix
from sklearn.utils.multiclass import (
    is_multilabel as sklearn_is_multilabel,
    type_of_target as sklearn_type_of_target,
)

from ... import opcodes as OperandDef
from ... import tensor as mt
from ...core import ENTITY_TYPE, TILEABLE_TYPE, recursive_tile
from ...core.context import get_context
from ...serialization.serializables import AnyField, ListField
from ...tensor.core import TensorOrder, TENSOR_TYPE
from ...typing import TileableType
from ...utils import has_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin, OutputType
from ..utils import assert_all_finite
from .validation import check_array


def _unique_multiclass(y):
    if hasattr(y, "__array__") or hasattr(y, "__mars_tensor__"):
        return mt.unique(mt.asarray(y))
    else:
        return set(y)


def _unique_indicator(y):
    return mt.arange(check_array(y, accept_sparse=True).shape[1])


_FN_UNIQUE_LABELS = {
    "binary": _unique_multiclass,
    "multiclass": _unique_multiclass,
    "multilabel-indicator": _unique_indicator,
}


class UniqueLabels(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.UNIQUE_LABELS

    ys = ListField("ys")

    def __call__(self, ys: List[TileableType]):
        self._output_types = [OutputType.tensor]
        inputs = [y for y in ys if isinstance(y, TILEABLE_TYPE)]
        return self.new_tileable(
            inputs,
            shape=(np.nan,),
            dtype=mt.tensor(ys[0]).dtype,
            order=TensorOrder.C_ORDER,
        )

    @classmethod
    def tile(cls, op: "UniqueLabels"):
        ys = op.ys
        ctx = get_context()

        target_types = yield from recursive_tile([type_of_target(x) for x in ys])
        # yield chunks of target_types for execution
        chunks = list(itertools.chain(*(t.chunks for t in target_types)))
        yield chunks

        ys_types = set(
            [it.item() for it in ctx.get_chunks_result([c.key for c in chunks])]
        )
        if ys_types == {"binary", "multiclass"}:
            ys_types = {"multiclass"}

        if len(ys_types) > 1:
            raise ValueError("Mix type of y not allowed, got types %s" % ys_types)

        label_type = ys_types.pop()

        # Check consistency for the indicator format
        if label_type == "multilabel-indicator":
            check_arrays = []
            chunks = []
            for y in ys:
                arr = yield from recursive_tile(check_array(y, accept_sparse=True))
                check_arrays.append(arr)
                chunks.extend(arr.chunks)
            yield check_arrays + chunks
            if len(set(arr.shape[1] for arr in check_arrays)) > 1:
                raise ValueError(
                    "Multi-label binary indicator input with "
                    "different numbers of labels"
                )

        # Get the unique set of labels
        _unique_labels = _FN_UNIQUE_LABELS.get(label_type, None)
        if not _unique_labels:
            raise ValueError("Unknown label type: %s" % repr(ys))

        labels = [_unique_labels(y) for y in ys]
        labels_chunks = []
        ys_labels = set()
        for label in labels:
            if isinstance(label, ENTITY_TYPE):
                label = yield from recursive_tile(label)
                labels_chunks.extend(label.chunks)
            else:
                ys_labels.update(label)
        yield labels_chunks
        ys_labels.update(
            itertools.chain.from_iterable(
                ctx.get_chunks_result([c.key for c in labels_chunks])
            )
        )

        # Check that we don't mix string type with number type
        if len(set(isinstance(label, str) for label in ys_labels)) > 1:
            raise ValueError("Mix of label input types (string and number)")

        return (yield from recursive_tile(mt.array(sorted(ys_labels))))


def unique_labels(*ys):
    """
    Extract an ordered array of unique labels.

    We don't allow:
        - mix of multilabel and multiclass (single label) targets
        - mix of label indicator matrix and anything else,
          because there are no explicit labels)
        - mix of label indicator matrices of different sizes
        - mix of string and integer labels

    At the moment, we also don't allow "multiclass-multioutput" input type.

    Parameters
    ----------
    *ys : array-likes

    Returns
    -------
    out : ndarray of shape (n_unique_labels,)
        An ordered array of unique labels.

    Examples
    --------
    >>> from mars.learn.utils.multiclass import unique_labels
    >>> unique_labels([3, 5, 5, 5, 7, 7]).execute()
    array([3, 5, 7])
    >>> unique_labels([1, 2, 3, 4], [2, 2, 3, 4]).execute()
    array([1, 2, 3, 4])
    >>> unique_labels([1, 2, 10], [5, 11]).execute()
    array([ 1,  2,  5, 10, 11])
    """
    if not ys:
        raise ValueError("No argument has been passed.")

    ys = list(ys)
    op = UniqueLabels(ys=ys)
    return op(ys)


class IsMultilabel(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.IS_MULTILABEL

    y = AnyField("y")

    def __call__(self, y):
        self._output_types = [OutputType.tensor]
        inputs = [y] if isinstance(y, ENTITY_TYPE) else []
        return self.new_tileable(
            inputs, shape=(), dtype=np.dtype(bool), order=TensorOrder.C_ORDER
        )

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self._inputs:
            self.y = self._inputs[0]

    @classmethod
    def _tile(cls, op: "IsMultilabel"):
        y = op.y

        if not isinstance(y, ENTITY_TYPE):
            return sklearn_is_multilabel(y)

        ctx = get_context()

        if has_unknown_shape(y):  # pragma: no cover
            yield

        if not (hasattr(y, "shape") and y.ndim == 2 and y.shape[1] > 1):
            return False

        labels = yield from recursive_tile(mt.unique(y))
        yield labels.chunks + [labels]

        if len(labels) < 3:
            if y.dtype.kind in "biu":
                return True
            if y.dtype.kind == "f":
                is_integral_float = yield from recursive_tile(
                    mt.all(mt.equal(y.astype(int), y))
                )
                yield is_integral_float.chunks
                is_integral_float = ctx.get_chunks_result(
                    [is_integral_float.chunks[0].key]
                )[0]
                if is_integral_float:
                    return True

        return False

    @classmethod
    def tile(cls, op: "IsMultilabel"):
        result = yield from cls._tile(op)
        return (yield from recursive_tile(mt.array(result)))


def is_multilabel(y):
    """
    Check if ``y`` is in a multilabel format.

    Parameters
    ----------
    y : numpy array of shape [n_samples]
        Target values.

    Returns
    -------
    out : bool,
        Return ``True``, if ``y`` is in a multilabel format, else ```False``.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.learn.utils.multiclass import is_multilabel
    >>> is_multilabel([0, 1, 0, 1]).execute()
    False
    >>> is_multilabel([[1], [0, 2], []]).execute()
    False
    >>> is_multilabel(mt.array([[1, 0], [0, 0]])).execute()
    True
    >>> is_multilabel(mt.array([[1], [0], [0]])).execute()
    False
    >>> is_multilabel(mt.array([[1, 0, 0]])).execute()
    True
    """
    if not isinstance(y, ENTITY_TYPE):
        if hasattr(y, "__array__") or isinstance(y, Sequence):
            y = np.asarray(y)
        yt = None
    else:
        yt = y = mt.tensor(y)

    op = IsMultilabel(y=y)
    return op(yt)


class TypeOfTarget(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.TYPE_OF_TARGET

    y = AnyField("y")

    def __call__(self, y: TileableType):
        self._output_types = [OutputType.tensor]
        inputs = [y] if isinstance(y, ENTITY_TYPE) else []
        return self.new_tileable(
            inputs, shape=(), order=TensorOrder.C_ORDER, dtype=np.dtype(object)
        )

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self._inputs:
            self.y = self._inputs[0]

    @classmethod
    def _tile(cls, op: "TypeOfTarget"):
        y = op.y

        # y is ndarray
        if not isinstance(y, ENTITY_TYPE):
            return sklearn_type_of_target(y)
        else:
            # make sure y executed
            yield

        ctx = get_context()

        multilabel = yield from recursive_tile(is_multilabel(y))
        yield multilabel.chunks
        multilabel = ctx.get_chunks_result([multilabel.chunks[0].key])[0]
        if multilabel:
            return "multilabel-indicator"

        # Invalid inputs
        if y.ndim > 2:
            return "unknown"
        if y.dtype == object and len(y):
            # [[[1, 2]]] or [obj_1] and not ["label_1"]
            first_val = ctx.get_chunks_result([y.chunks[0].key])[0].flat[0]
            if not isinstance(first_val, str):
                return "unknown"

        if y.ndim == 2 and y.shape[1] == 0:
            return "unknown"  # [[]]

        if y.ndim == 2 and y.shape[1] > 1:
            suffix = "-multioutput"  # [[1, 2], [1, 2]]
        else:
            suffix = ""  # [1, 2, 3] or [[1], [2], [3]]

        if y.dtype.kind == "f":
            # check float and contains non-integer float values
            contain_float_values = yield from recursive_tile(mt.any(y != y.astype(int)))
            yield contain_float_values.chunks
            contain_float_values = ctx.get_chunks_result(
                [contain_float_values.chunks[0].key]
            )[0]
            # [.1, .2, 3] or [[.1, .2, 3]] or [[1., .2]] and not [1., 2., 3.]
            if contain_float_values:
                yield from recursive_tile(assert_all_finite(y))
                return "continuous" + suffix

        unique_y = yield from recursive_tile(mt.unique(y))
        yield unique_y.chunks + [unique_y]
        if (len(unique_y) > 2) or (y.ndim >= 2 and len(y[0]) > 1):
            return "multiclass" + suffix  # [1, 2, 3] or [[1., 2., 3]] or [[1, 2]]
        else:
            return "binary"  # [1, 2] or [["a"], ["b"]]

    @classmethod
    def tile(cls, op: "TypeOfTarget"):
        result = yield from cls._tile(op)
        return (yield from recursive_tile(mt.array(result)))


def type_of_target(y):
    """
    Determine the type of data indicated by the target.

    Note that this type is the most specific type that can be inferred.
    For example:

        * ``binary`` is more specific but compatible with ``multiclass``.
        * ``multiclass`` of integers is more specific but compatible with
          ``continuous``.
        * ``multilabel-indicator`` is more specific but compatible with
          ``multiclass-multioutput``.

    Parameters
    ----------
    y : array-like

    Returns
    -------
    target_type : string
        One of:

        * 'continuous': `y` is an array-like of floats that are not all
          integers, and is 1d or a column vector.
        * 'continuous-multioutput': `y` is a 2d tensor of floats that are
          not all integers, and both dimensions are of size > 1.
        * 'binary': `y` contains <= 2 discrete values and is 1d or a column
          vector.
        * 'multiclass': `y` contains more than two discrete values, is not a
          sequence of sequences, and is 1d or a column vector.
        * 'multiclass-multioutput': `y` is a 2d tensor that contains more
          than two discrete values, is not a sequence of sequences, and both
          dimensions are of size > 1.
        * 'multilabel-indicator': `y` is a label indicator matrix, a tensor
          of two dimensions with at least two columns, and at most 2 unique
          values.
        * 'unknown': `y` is array-like but none of the above, such as a 3d
          tensor, sequence of sequences, or a tensor of non-sequence objects.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.learn.utils.multiclass import type_of_target
    >>> type_of_target([0.1, 0.6]).execute()
    'continuous'
    >>> type_of_target([1, -1, -1, 1]).execute()
    'binary'
    >>> type_of_target(['a', 'b', 'a']).execute()
    'binary'
    >>> type_of_target([1.0, 2.0]).execute()
    'binary'
    >>> type_of_target([1, 0, 2]).execute()
    'multiclass'
    >>> type_of_target([1.0, 0.0, 3.0]).execute()
    'multiclass'
    >>> type_of_target(['a', 'b', 'c']).execute()
    'multiclass'
    >>> type_of_target(mt.array([[1, 2], [3, 1]])).execute()
    'multiclass-multioutput'
    >>> type_of_target([[1, 2]]).execute()
    'multiclass-multioutput'
    >>> type_of_target(mt.array([[1.5, 2.0], [3.0, 1.6]])).execute()
    'continuous-multioutput'
    >>> type_of_target(mt.array([[0, 1], [1, 1]])).execute()
    'multilabel-indicator'
    """
    if isinstance(y, TENSOR_TYPE):
        y = mt.tensor(y)

    valid_types = (Sequence, spmatrix) if spmatrix is not None else (Sequence,)
    valid = (
        isinstance(y, valid_types)
        or hasattr(y, "__array__")
        or hasattr(y, "__mars_tensor__")
    ) and not isinstance(y, str)

    if not valid:
        raise ValueError(f"Expected array-like (array or non-string sequence), got {y}")

    sparse_pandas = y.__class__.__name__ in ["SparseSeries", "SparseArray"]
    if sparse_pandas:  # pragma: no cover
        raise ValueError("y cannot be class 'SparseSeries' or 'SparseArray'")

    if isinstance(y, ENTITY_TYPE):
        y = mt.tensor(y)

    op = TypeOfTarget(y=y)
    return op(y)


def check_classification_targets(y):
    """
    Ensure that target y is of a non-regression type.

    Only the following target types (as defined in type_of_target) are allowed:
        'binary', 'multiclass', 'multiclass-multioutput',
        'multilabel-indicator', 'multilabel-sequences'

    Parameters
    ----------
    y : array-like
    """
    y_type = type_of_target(y)

    def check(t):
        if t not in [
            "binary",
            "multiclass",
            "multiclass-multioutput",
            "multilabel-indicator",
            "multilabel-sequences",
        ]:
            raise ValueError("Unknown label type: %r" % t)
        return t

    y_type = y_type.map_chunk(check, dtype=y_type.dtype)
    return y_type



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Dict, Union

from ...resource import Resource
from ...services import start_services, stop_services, NodeRole

logger = logging.getLogger(__name__)


async def start_supervisor(
    address: str,
    lookup_address: str = None,
    modules: Union[List, str, None] = None,
    config: Dict = None,
    web: Union[str, bool] = "auto",
):
    logger.debug("Starting Mars supervisor at %s", address)
    lookup_address = lookup_address or address
    backend = config["cluster"].get("backend", "fixed")
    if backend == "fixed" and config["cluster"].get("lookup_address") is None:
        config["cluster"]["lookup_address"] = lookup_address
    if web:
        # try to append web to services
        config["services"].append("web")
    if modules:
        config["modules"] = modules
    try:
        await start_services(NodeRole.SUPERVISOR, config, address=address)
        logger.debug("Mars supervisor started at %s", address)
    except ImportError:
        if web == "auto":
            config["services"] = [
                service for service in config["services"] if service != "web"
            ]
            await start_services(NodeRole.SUPERVISOR, config, address=address)
            logger.debug("Mars supervisor started at %s", address)
            return False
        else:  # pragma: no cover
            raise
    else:
        return bool(web)


async def stop_supervisor(address: str, config: Dict = None):
    await stop_services(NodeRole.SUPERVISOR, address=address, config=config)


async def start_worker(
    address: str,
    lookup_address: str,
    band_to_resource: Dict[str, Resource],
    modules: Union[List, str, None] = None,
    config: Dict = None,
    mark_ready: bool = True,
):
    logger.debug("Starting Mars worker at %s", address)
    backend = config["cluster"].get("backend", "fixed")
    if backend == "fixed" and config["cluster"].get("lookup_address") is None:
        config["cluster"]["lookup_address"] = lookup_address
    if config["cluster"].get("resource") is None:
        config["cluster"]["resource"] = band_to_resource
    if any(
        band_name.startswith("gpu-") for band_name in band_to_resource
    ):  # pragma: no cover
        if "cuda" not in config["storage"]["backends"]:
            config["storage"]["backends"].append("cuda")
    if modules:
        config["modules"] = modules
    await start_services(
        NodeRole.WORKER, config, address=address, mark_ready=mark_ready
    )
    logger.debug("Mars worker started at %s", address)


async def stop_worker(address: str, config: Dict = None):
    await stop_services(NodeRole.WORKER, address=address, config=config)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dask.core import istask, ishashable

from typing import List, Tuple, Union
from .utils import reduce
from ...remote import spawn
from ...deploy.oscar.session import execute


def mars_scheduler(dsk: dict, keys: Union[List[List[str]], List[str]]):
    """
    A Dask-Mars scheduler

    This scheduler is intended to be compatible with existing
    dask user interface, no callbacks are implemented.

    Parameters
    ----------
    dsk: Dict
        Dask graph, represented as a task DAG dictionary.
    keys: Union[List[List[str]], List[str]]
        1d or 2d list of Dask graph keys whose values we wish to compute and return.

    Returns
    -------
    Object
        Computed values corresponding to the provided keys with same dimension.
    """

    if isinstance(keys, List) and not isinstance(keys[0], List):  # 1d keys
        task = execute(mars_dask_get(dsk, keys))
        if not isinstance(task, List):
            task = [task]
        return map(lambda x: x.fetch(), task)
    else:  # 2d keys
        res = execute(reduce(mars_dask_get(dsk, keys))).fetch()
        if not isinstance(res, List):
            return [[res]]
        else:
            return res


def mars_dask_get(dsk: dict, keys: Union[List[List[str]], List[str]]):
    """
    A Dask-Mars convert function. This function will send the dask graph layers
        to Mars Remote API, generating mars objects correspond to the provided keys.

    Parameters
    ----------
    dsk: Dict
        Dask graph, represented as a task DAG dictionary.
    keys: Union[List[List[str]], List[str]]
        1d or 2d list of Dask graph keys whose values we wish to compute and return.

    Returns
    -------
    Object
        Spawned mars objects corresponding to the provided keys with same dimension.
    """

    def _get_arg(a):
        # if arg contains layer index or callable objs, handle it
        if ishashable(a) and a in dsk.keys():
            while ishashable(a) and a in dsk.keys():
                a = dsk[a]
            return _spawn_task(a)
        elif not isinstance(a, str) and hasattr(a, "__getitem__"):
            if istask(
                a
            ):  # TODO:Handle `SubgraphCallable`, which may contains dsk in it
                return spawn(a[0], args=tuple(_get_arg(i) for i in a[1:]))
            elif isinstance(a, dict):
                return {k: _get_arg(v) for k, v in a.items()}
            elif isinstance(a, List) or isinstance(a, Tuple):
                return type(a)(_get_arg(i) for i in a)
        return a

    def _spawn_task(task: tuple):
        if not istask(task):
            return _get_arg(task)
        return spawn(task[0], args=tuple(_get_arg(a) for a in task[1:]))

    return [
        [_spawn_task(dsk[k]) for k in keys_d]
        if isinstance(keys_d, List)
        else _spawn_task(dsk[keys_d])
        for keys_d in keys
    ]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import numpy as np
from sklearn.base import ClusterMixin, TransformerMixin
from sklearn.exceptions import ConvergenceWarning

from ... import tensor as mt
from ...tensor.utils import check_random_state
from ..base import BaseEstimator
from ..metrics.pairwise import euclidean_distances
from ..utils.extmath import row_norms
from ..utils.validation import (
    _check_sample_weight,
    check_array,
    _num_samples,
    check_is_fitted,
)
from ._k_means_common import _inertia
from ._k_means_elkan_iter import init_bounds, elkan_iter
from ._k_means_init import _k_init, _scalable_k_init
from ._k_means_lloyd_iter import lloyd_iter


###############################################################################
# K-means batch estimation by EM (expectation maximization)


def _validate_center_shape(X, n_centers, centers):
    """Check if centers is compatible with X and n_centers"""
    if len(centers) != n_centers:
        raise ValueError(
            "The shape of the initial centers (%s) "
            "does not match the number of clusters %i" % (centers.shape, n_centers)
        )
    if centers.shape[1] != X.shape[1]:
        raise ValueError(
            "The number of features of the initial centers %s "
            "does not match the number of features of the data %s."
            % (centers.shape[1], X.shape[1])
        )


def _tolerance(X, tol):
    """Return a tolerance which is independent of the dataset"""
    variances = mt.var(X, axis=0)
    return mt.mean(variances) * tol


def _check_normalize_sample_weight(sample_weight, X):
    """Set sample_weight if None, and check for correct dtype"""

    sample_weight_was_none = sample_weight is None

    sample_weight = _check_sample_weight(sample_weight, X, dtype=X.dtype)
    if not sample_weight_was_none:
        # normalize the weights to sum up to n_samples
        # an array of 1 (i.e. samples_weight is None) is already normalized
        n_samples = len(sample_weight)
        scale = n_samples / sample_weight.sum()
        sample_weight *= scale
    return sample_weight


def k_means(
    X,
    n_clusters,
    sample_weight=None,
    init="k-means||",
    n_init=10,
    max_iter=300,
    verbose=False,
    tol=1e-4,
    random_state=None,
    copy_x=True,
    algorithm="auto",
    oversampling_factor=2,
    init_iter=5,
    return_n_iter=False,
):
    """K-means clustering algorithm.

    Parameters
    ----------
    X : Tensor, shape (n_samples, n_features)
        The observations to cluster. It must be noted that the data
        will be converted to C ordering, which will cause a memory copy
        if the given data is not C-contiguous.

    n_clusters : int
        The number of clusters to form as well as the number of
        centroids to generate.

    sample_weight : array-like, shape (n_samples,), optional
        The weights for each observation in X. If None, all observations
        are assigned equal weight (default: None)

    init : {'k-means++', 'k-means||', 'random', or tensor, or a callable}, optional
        Method for initialization, default to 'k-means||':

        'k-means++' : selects initial cluster centers for k-mean
        clustering in a smart way to speed up convergence. See section
        Notes in k_init for more details.

        'k-means||': scalable k-means++.

        'random': choose k observations (rows) at random from data for
        the initial centroids.

        If an ndarray is passed, it should be of shape (n_clusters, n_features)
        and gives the initial centers.

        If a callable is passed, it should take arguments X, k and
        and a random state and return an initialization.

    n_init : int, optional, default: 10
        Number of time the k-means algorithm will be run with different
        centroid seeds. The final results will be the best output of
        n_init consecutive runs in terms of inertia.

    max_iter : int, optional, default 300
        Maximum number of iterations of the k-means algorithm to run.

    verbose : boolean, optional
        Verbosity mode.

    tol : float, optional
        The relative increment in the results before declaring convergence.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for centroid initialization. Use
        an int to make the randomness deterministic.
        See :term:`Glossary <random_state>`.

    copy_x : bool, optional
        When pre-computing distances it is more numerically accurate to center
        the data first.  If copy_x is True (default), then the original data is
        not modified, ensuring X is C-contiguous.  If False, the original data
        is modified, and put back before the function returns, but small
        numerical differences may be introduced by subtracting and then adding
        the data mean, in this case it will also not ensure that data is
        C-contiguous which may cause a significant slowdown.

    algorithm : "auto", "full" or "elkan", default="auto"
        K-means algorithm to use. The classical EM-style algorithm is "full".
        The "elkan" variation is more efficient by using the triangle
        inequality, but currently doesn't support sparse data. "auto" chooses
        "elkan" for dense data and "full" for sparse data.

    oversampling_factor: int, default=2
        Only work for kmeans||, used in each iteration in kmeans||.

    init_iter: int, default=5
        Only work for kmeans||, indicates how may iterations required.

    return_n_iter : bool, optional
        Whether or not to return the number of iterations.

    Returns
    -------
    centroid : float ndarray with shape (k, n_features)
        Centroids found at the last iteration of k-means.

    label : integer ndarray with shape (n_samples,)
        label[i] is the code or index of the centroid the
        i'th observation is closest to.

    inertia : float
        The final value of the inertia criterion (sum of squared distances to
        the closest centroid for all observations in the training set).

    best_n_iter : int
        Number of iterations corresponding to the best results.
        Returned only if `return_n_iter` is set to True.
    """

    est = KMeans(
        n_clusters=n_clusters,
        init=init,
        n_init=n_init,
        max_iter=max_iter,
        verbose=verbose,
        tol=tol,
        random_state=random_state,
        copy_x=copy_x,
        algorithm=algorithm,
        oversampling_factor=oversampling_factor,
        init_iter=init_iter,
    ).fit(X, sample_weight=sample_weight)
    if return_n_iter:
        return est.cluster_centers_, est.labels_, est.inertia_, est.n_iter_
    else:
        return est.cluster_centers_, est.labels_, est.inertia_


def _kmeans_single_elkan(
    X,
    sample_weight,
    centers_init,
    n_clusters,
    max_iter=300,
    verbose=False,
    x_squared_norms=None,
    tol=1e-4,
    X_mean=None,
    session=None,
    run_kwargs=None,
):
    sample_weight = _check_normalize_sample_weight(sample_weight, X)

    centers = centers_init
    # execute X, centers and tol first
    tol = mt.asarray(tol)
    to_run = [X, sample_weight, centers, x_squared_norms, tol]
    if X_mean is not None:
        to_run.append(X_mean)
    mt.ExecutableTuple(to_run).execute(session=session, **(run_kwargs or dict()))
    tol = tol.fetch(session=session)

    if verbose:
        print("Initialization complete")

    center_half_distances = euclidean_distances(centers) / 2
    distance_next_center = mt.partition(
        mt.asarray(center_half_distances), kth=1, axis=0
    )[1]
    center_shift = mt.zeros(n_clusters, dtype=X.dtype)

    labels, upper_bounds, lower_bounds = init_bounds(
        X, centers, center_half_distances, n_clusters
    )

    for i in range(max_iter):
        to_runs = []

        (
            centers_new,
            weight_in_clusters,
            upper_bounds,
            lower_bounds,
            labels,
            center_shift,
        ) = elkan_iter(
            X,
            sample_weight,
            centers,
            center_half_distances,
            distance_next_center,
            upper_bounds,
            lower_bounds,
            labels,
            center_shift,
            session=session,
            run_kwargs=run_kwargs,
        )
        to_runs.extend(
            [
                centers_new,
                weight_in_clusters,
                upper_bounds,
                lower_bounds,
                labels,
                center_shift,
            ]
        )

        # compute new pairwise distances between centers and closest other
        # center of each center for next iterations
        center_half_distances = euclidean_distances(centers_new) / 2
        distance_next_center = mt.partition(
            mt.asarray(center_half_distances), kth=1, axis=0
        )[1]
        to_runs.extend([center_half_distances, distance_next_center])

        if verbose:
            inertia = _inertia(X, sample_weight, centers, labels)
            to_runs.append(inertia)

        center_shift_tot = (center_shift**2).sum()
        to_runs.append(center_shift_tot)

        mt.ExecutableTuple(to_runs).execute(session=session, **(run_kwargs or dict()))

        if verbose:
            inertia_data = inertia.fetch(session=session)
            print(f"Iteration {i}, inertia {inertia_data}")

        center_shift_tot = center_shift_tot.fetch(session=session)
        if center_shift_tot <= tol:
            if verbose:  # pragma: no cover
                print(
                    f"Converged at iteration {i}: center shift {center_shift_tot} "
                    f"within tolerance {tol}"
                )
            break

        centers, centers_new = centers_new, centers

    if center_shift_tot > 0:
        # rerun E-step so that predicted labels match cluster centers
        (
            centers_new,
            weight_in_clusters,
            upper_bounds,
            lower_bounds,
            labels,
            center_shift,
        ) = elkan_iter(
            X,
            sample_weight,
            centers,
            center_half_distances,
            distance_next_center,
            upper_bounds,
            lower_bounds,
            labels,
            center_shift,
            update_centers=False,
            session=session,
            run_kwargs=run_kwargs,
        )

    inertia = _inertia(X, sample_weight, centers, labels)

    mt.ExecutableTuple([labels, inertia, centers]).execute(
        session=session, **(run_kwargs or dict())
    )
    return labels, inertia, centers, i + 1


def _kmeans_single_lloyd(
    X,
    sample_weight,
    centers_init,
    n_clusters,
    max_iter=300,
    verbose=False,
    x_squared_norms=None,
    tol=1e-4,
    X_mean=None,
    session=None,
    run_kwargs=None,
):
    sample_weight = _check_normalize_sample_weight(sample_weight, X)

    centers = centers_init
    # execute X, centers and tol first
    tol = mt.asarray(tol)
    to_run = [X, centers, x_squared_norms, tol]
    if X_mean is not None:
        to_run.append(X_mean)
    mt.ExecutableTuple(to_run).execute(session=session, **(run_kwargs or dict()))
    tol = tol.fetch(session=session)

    if verbose:  # pragma: no cover
        print("Initialization complete")

    labels = mt.full(X.shape[0], -1, dtype=mt.int32)
    center_shift = mt.zeros(n_clusters, dtype=X.dtype)

    for i in range(max_iter):
        to_runs = []

        centers_new, weight_in_clusters, labels, center_shift = lloyd_iter(
            X,
            sample_weight,
            x_squared_norms,
            centers,
            labels,
            center_shift,
            update_centers=True,
            session=session,
            run_kwargs=run_kwargs,
        )
        to_runs.extend([centers_new, weight_in_clusters, labels, center_shift])

        if verbose:
            inertia = _inertia(X, sample_weight, centers, labels)
            to_runs.append(inertia)

        center_shift_tot = (center_shift**2).sum()
        to_runs.append(center_shift_tot)

        mt.ExecutableTuple(to_runs).execute(session=session, **(run_kwargs or dict()))

        if verbose:  # pragma: no cover
            inertia_data = inertia.fetch(session=session)
            print(f"Iteration {i}, inertia {inertia_data}")

        center_shift_tot = center_shift_tot.fetch(session=session)
        if center_shift_tot <= tol:
            if verbose:  # pragma: no cover
                print(
                    f"Converged at iteration {i}: center shift {center_shift_tot} "
                    f"within tolerance {tol}"
                )
            break

        centers, centers_new = centers_new, centers

    if center_shift_tot > 0:
        # rerun E-step so that predicted labels match cluster centers
        centers_new, weight_in_clusters, labels, center_shift = lloyd_iter(
            X,
            sample_weight,
            x_squared_norms,
            centers,
            labels,
            center_shift,
            update_centers=False,
            session=session,
            run_kwargs=run_kwargs,
        )

    inertia = _inertia(X, sample_weight, centers, labels)

    mt.ExecutableTuple([labels, inertia, centers]).execute(
        session=session, **(run_kwargs or dict())
    )
    return labels, inertia, centers, i + 1


def _labels_inertia(
    X, sample_weight, x_squared_norms, centers, session=None, run_kwargs=None
):
    """E step of the K-means EM algorithm.

    Compute the labels and the inertia of the given samples and centers.

    Parameters
    ----------
    X : {array-like, sparse matrix} of shape (n_samples, n_features)
        The input samples to assign to the labels. If sparse matrix, must be in
        CSR format.

    sample_weight : array-like of shape (n_samples,)
        The weights for each observation in X.

    x_squared_norms : Tensor of shape (n_samples,)
        Precomputed squared euclidean norm of each data point, to speed up
        computations.

    centers : Tensor, shape (n_clusters, n_features)
        The cluster centers.

    Returns
    -------
    labels : ndarray of shape (n_samples,)
        The resulting assignment

    inertia : float
        Sum of squared distances of samples to their closest cluster center.
    """
    n_samples = X.shape[0]
    n_clusters = centers.shape[0]

    sample_weight = _check_normalize_sample_weight(sample_weight, X)
    labels = mt.full(n_samples, -1, dtype=np.int32)
    weight_in_clusters = mt.zeros(n_clusters, dtype=centers.dtype)
    center_shift = mt.zeros_like(weight_in_clusters)

    centers, weight_in_clusters, labels, center_shift = lloyd_iter(
        X,
        sample_weight,
        x_squared_norms,
        centers,
        labels,
        center_shift,
        update_centers=False,
        session=session,
        run_kwargs=run_kwargs,
    )

    inertia = _inertia(X, sample_weight, centers, labels)

    return labels, inertia


def _init_centroids(
    X,
    n_clusters=8,
    init="k-means++",
    random_state=None,
    x_squared_norms=None,
    init_size=None,
    oversampling_factor=2,
    init_iter=5,
):
    """Compute the initial centroids

    Parameters
    ----------

    X : Tensor of shape (n_samples, n_features)
        The input samples.

    n_clusters : int, default=8
        number of centroids.

    init : {'k-means++', 'k-means||', 'random', tensor, callable}, default="k-means++"
        Method for initialization.

    random_state : int, RandomState instance, default=None
        Determines random number generation for centroid initialization. Use
        an int to make the randomness deterministic.
        See :term:`Glossary <random_state>`.

    x_squared_norms : tensor of shape (n_samples,), default=None
        Squared euclidean norm of each data point. Pass it if you have it at
        hands already to avoid it being recomputed here. Default: None

    init_size : int, default=None
        Number of samples to randomly sample for speeding up the
        initialization (sometimes at the expense of accuracy): the
        only algorithm is initialized by running a batch KMeans on a
        random subset of the data. This needs to be larger than k.

    Returns
    -------
    centers : tensor of shape(k, n_features)
    """
    random_state = check_random_state(random_state).to_numpy()
    n_samples = X.shape[0]

    if x_squared_norms is None:
        x_squared_norms = row_norms(X, squared=True)

    if init_size is not None and init_size < n_samples:  # pragma: no cover
        if init_size < n_clusters:
            warnings.warn(
                f"init_size={init_size} should be larger than k={n_clusters}. "
                "Setting it to 3*k",
                RuntimeWarning,
                stacklevel=2,
            )
            init_size = 3 * n_clusters
        init_indices = random_state.randint(0, n_samples, init_size)
        X = X[init_indices]
        x_squared_norms = x_squared_norms[init_indices]
        n_samples = X.shape[0]
    elif n_samples < n_clusters:
        raise ValueError(
            f"n_samples={n_samples} should be larger than n_clusters={n_clusters}"
        )

    if isinstance(init, str) and init == "k-means++":
        centers = _k_init(
            X, n_clusters, random_state=random_state, x_squared_norms=x_squared_norms
        )
    elif isinstance(init, str) and init == "k-means||":
        centers = _scalable_k_init(
            X,
            n_clusters,
            random_state=random_state,
            x_squared_norms=x_squared_norms,
            oversampling_factor=oversampling_factor,
            init_iter=init_iter,
        )
    elif isinstance(init, str) and init == "random":
        seeds = random_state.choice(n_samples, size=n_clusters, replace=False)
        centers = X[seeds].rechunk((n_clusters, X.shape[1]))
    elif hasattr(init, "__array__"):
        # ensure that the centers have the same dtype as X
        # this is a requirement of fused types of cython
        centers = mt.array(init, dtype=X.dtype)
    elif callable(init):
        centers = init(X, n_clusters, random_state=random_state)
        centers = mt.asarray(centers, dtype=X.dtype)
    else:  # pragma: no cover
        raise ValueError(
            "the init parameter for the k-means should "
            "be 'k-means++' or 'random' or a tensor, "
            f"'{init}' (type '{type(init)}') was passed."
        )

    if centers.issparse():
        centers = centers.todense()

    _validate_center_shape(X, n_clusters, centers)
    return centers


class KMeans(TransformerMixin, ClusterMixin, BaseEstimator):
    """K-Means clustering.

    Read more in the :ref:`User Guide <k_means>`.

    Parameters
    ----------

    n_clusters : int, default=8
        The number of clusters to form as well as the number of
        centroids to generate.

    init : {'k-means++', 'k-means||', 'random'} or tensor of shape \
            (n_clusters, n_features), default='k-means||'
        Method for initialization, defaults to 'k-means||':

        'k-means++' : selects initial cluster centers for k-mean
        clustering in a smart way to speed up convergence. See section
        Notes in k_init for more details.

        'k-means||': scalable k-means++.

        'random': choose k observations (rows) at random from data for
        the initial centroids.

        If a tensor is passed, it should be of shape (n_clusters, n_features)
        and gives the initial centers.

    n_init : int, default=1
        Number of time the k-means algorithm will be run with different
        centroid seeds. The final results will be the best output of
        n_init consecutive runs in terms of inertia.

    max_iter : int, default=300
        Maximum number of iterations of the k-means algorithm for a
        single run.

    tol : float, default=1e-4
        Relative tolerance with regards to inertia to declare convergence.

    verbose : int, default=0
        Verbosity mode.

    random_state : int, RandomState instance, default=None
        Determines random number generation for centroid initialization. Use
        an int to make the randomness deterministic.
        See :term:`Glossary <random_state>`.

    copy_x : bool, default=True
        When pre-computing distances it is more numerically accurate to center
        the data first.  If copy_x is True (default), then the original data is
        not modified, ensuring X is C-contiguous.  If False, the original data
        is modified, and put back before the function returns, but small
        numerical differences may be introduced by subtracting and then adding
        the data mean, in this case it will also not ensure that data is
        C-contiguous which may cause a significant slowdown.

    algorithm : {"auto", "full", "elkan"}, default="auto"
        K-means algorithm to use. The classical EM-style algorithm is "full".
        The "elkan" variation is more efficient by using the triangle
        inequality, but currently doesn't support sparse data. "auto" chooses
        "elkan" for dense data and "full" for sparse data.

    oversampling_factor: int, default=2
        Only work for kmeans||, used in each iteration in kmeans||.

    init_iter: int, default=5
        Only work for kmeans||, indicates how may iterations required.

    Attributes
    ----------
    cluster_centers_ : tensor of shape (n_clusters, n_features)
        Coordinates of cluster centers. If the algorithm stops before fully
        converging (see ``tol`` and ``max_iter``), these will not be
        consistent with ``labels_``.

    labels_ : tensor of shape (n_samples,)
        Labels of each point

    inertia_ : float
        Sum of squared distances of samples to their closest cluster center.

    n_iter_ : int
        Number of iterations run.

    See Also
    --------

    MiniBatchKMeans
        Alternative online implementation that does incremental updates
        of the centers positions using mini-batches.
        For large scale learning (say n_samples > 10k) MiniBatchKMeans is
        probably much faster than the default batch implementation.

    Notes
    -----
    The k-means problem is solved using either Lloyd's or Elkan's algorithm.

    The average complexity is given by O(k n T), were n is the number of
    samples and T is the number of iteration.

    The worst case complexity is given by O(n^(k+2/p)) with
    n = n_samples, p = n_features. (D. Arthur and S. Vassilvitskii,
    'How slow is the k-means method?' SoCG2006)

    In practice, the k-means algorithm is very fast (one of the fastest
    clustering algorithms available), but it falls in local minima. That's why
    it can be useful to restart it several times.

    If the algorithm stops before fully converging (because of ``tol`` or
    ``max_iter``), ``labels_`` and ``cluster_centers_`` will not be consistent,
    i.e. the ``cluster_centers_`` will not be the means of the points in each
    cluster. Also, the estimator will reassign ``labels_`` after the last
    iteration to make ``labels_`` consistent with ``predict`` on the training
    set.

    Examples
    --------

    >>> from mars.learn.cluster import KMeans
    >>> import mars.tensor as mt
    >>> X = mt.array([[1, 2], [1, 4], [1, 0],
    ...               [10, 2], [10, 4], [10, 0]])
    >>> kmeans = KMeans(n_clusters=2, random_state=0, init='k-means++').fit(X)
    >>> kmeans.labels_
    array([1, 1, 1, 0, 0, 0], dtype=int32)
    >>> kmeans.predict([[0, 0], [12, 3]])
    array([1, 0], dtype=int32)
    >>> kmeans.cluster_centers_
    array([[10.,  2.],
           [ 1.,  2.]])
    """

    def __init__(
        self,
        n_clusters=8,
        init="k-means||",
        n_init=1,
        max_iter=300,
        tol=1e-4,
        verbose=0,
        random_state=None,
        copy_x=True,
        algorithm="auto",
        oversampling_factor=2,
        init_iter=5,
    ):

        self.n_clusters = n_clusters
        self.init = init
        self.max_iter = max_iter
        self.tol = tol
        self.n_init = n_init
        self.verbose = verbose
        self.random_state = random_state
        self.copy_x = copy_x
        self.algorithm = algorithm
        self.oversampling_factor = oversampling_factor
        self.init_iter = init_iter

    def _check_params(self, X):
        # n_init
        if self.n_init <= 0:
            raise ValueError(f"n_init should be > 0, got {self.n_init} instead.")
        self._n_init = self.n_init

        # max_iter
        if self.max_iter <= 0:
            raise ValueError(f"max_iter should be > 0, got {self.max_iter} instead.")

        # n_clusters
        if X.shape[0] < self.n_clusters:
            raise ValueError(
                f"n_samples={X.shape[0]} should be >= n_clusters={self.n_clusters}."
            )

        # tol
        self._tol = _tolerance(X, self.tol)

        # algorithm
        if self.algorithm not in ("auto", "full", "elkan"):
            raise ValueError(
                f"Algorithm must be 'auto', 'full' or 'elkan', "
                f"got {self.algorithm} instead."
            )

        self._algorithm = self.algorithm
        if self._algorithm == "auto":
            # note(xuye.qin):
            # Different from scikit-learn,
            # for now, full seems more efficient when data is large,
            # elkan needs to be tuned more
            # old: algorithm = "full" if self.n_clusters == 1 else "elkan"
            self._algorithm = "full"
        if self._algorithm == "elkan" and self.n_clusters == 1:
            warnings.warn(
                "algorithm='elkan' doesn't make sense for a single "
                "cluster. Using 'full' instead.",
                RuntimeWarning,
            )
            self._algorithm = "full"

        # init
        if not (
            hasattr(self.init, "__array__")
            or callable(self.init)
            or (
                isinstance(self.init, str)
                and self.init in ["k-means++", "k-means||", "random"]
            )
        ):
            raise ValueError(
                f"init should be either 'k-means++'， 'k-mean||', 'random', "
                f"a tensor, a ndarray or a "
                f"callable, got '{self.init}' instead."
            )

        if hasattr(self.init, "__array__") and self._n_init != 1:
            warnings.warn(
                f"Explicit initial center position passed: performing only"
                f" one init in {self.__class__.__name__} instead of "
                f"n_init={self._n_init}.",
                RuntimeWarning,
                stacklevel=2,
            )
            self._n_init = 1

    def _check_test_data(self, X):
        X = check_array(
            X,
            accept_sparse=True,
            dtype=[np.float64, np.float32],
            order="C",
            accept_large_sparse=False,
        )
        n_samples, n_features = X.shape
        expected_n_features = self.cluster_centers_.shape[1]
        if not n_features == expected_n_features:  # pragma: no cover
            raise ValueError(
                f"Incorrect number of features. Got {n_features} features, "
                f"expected {expected_n_features}"
            )

        return X

    def fit(self, X, y=None, sample_weight=None, session=None, run_kwargs=None):
        """Compute k-means clustering.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training instances to cluster. It must be noted that the data
            will be converted to C ordering, which will cause a memory
            copy if the given data is not C-contiguous.
            If a sparse matrix is passed, a copy will be made if it's not in
            CSR format.

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

        Returns
        -------
        self
            Fitted estimator.
        """
        expect_chunk_size_on_columns = mt.tensor(X).shape[1]
        if not np.isnan(expect_chunk_size_on_columns):
            X = mt.tensor(X, chunk_size={1: expect_chunk_size_on_columns})

        X = self._validate_data(
            X,
            accept_sparse=True,
            dtype=[np.float64, np.float32],
            order="C",
            copy=self.copy_x,
            accept_large_sparse=False,
        )
        # verify that the number of samples given is larger than k
        if np.isnan(_num_samples(X)):  # pragma: no cover
            X.execute(session=session, **(run_kwargs or dict()))

        self._check_params(X)
        random_state = check_random_state(self.random_state).to_numpy()

        tol = _tolerance(X, self.tol)

        # Validate init array
        init = self.init
        if hasattr(init, "__array__"):
            init = check_array(init, dtype=X.dtype.type, copy=True, order="C")
            _validate_center_shape(X, self.n_clusters, init)

        # subtract of mean of x for more accurate distance computations
        X_mean = None
        if not X.issparse():
            X_mean = X.mean(axis=0)
            # The copy was already done above
            X -= X_mean

            if hasattr(init, "__array__"):
                init -= X_mean

        # precompute squared norms of data points
        x_squared_norms = row_norms(X, squared=True)

        best_labels, best_inertia, best_centers = None, None, None

        if self._algorithm == "full":
            kmeans_single = _kmeans_single_lloyd
        else:
            kmeans_single = _kmeans_single_elkan

        for i in range(self._n_init):  # pylint: disable=unused-variable
            # Initialize centers
            centers_init = _init_centroids(
                X,
                self.n_clusters,
                init,
                random_state=random_state,
                x_squared_norms=x_squared_norms,
                oversampling_factor=self.oversampling_factor,
                init_iter=self.init_iter,
            )

            # run a k-means once
            labels, inertia, centers, n_iter_ = kmeans_single(
                X,
                sample_weight,
                centers_init,
                self.n_clusters,
                max_iter=self.max_iter,
                verbose=self.verbose,
                tol=tol,
                x_squared_norms=x_squared_norms,
                X_mean=X_mean,
                session=session,
                run_kwargs=run_kwargs,
            )
            inertia = inertia.fetch(session=session)
            # determine if these results are the best so far
            if best_inertia is None or inertia < best_inertia:
                best_labels = labels
                best_centers = centers
                best_inertia = inertia
                best_n_iter = n_iter_

        if not X.issparse():
            if not self.copy_x:  # pragma: no cover
                X += X_mean
            best_centers += X_mean
            best_centers.execute(session=session, **(run_kwargs or dict()))

        distinct_clusters = len(set(best_labels.fetch(session=session)))
        if distinct_clusters < self.n_clusters:  # pragma: no cover
            warnings.warn(
                f"Number of distinct clusters ({distinct_clusters}) found smaller than "
                f"n_clusters ({self.n_clusters}). Possibly due to duplicate points in X.",
                ConvergenceWarning,
                stacklevel=2,
            )

        self.cluster_centers_ = best_centers
        self.labels_ = best_labels
        self.inertia_ = best_inertia
        self.n_iter_ = best_n_iter
        return self

    def fit_predict(self, X, y=None, sample_weight=None, session=None, run_kwargs=None):
        """Compute cluster centers and predict cluster index for each sample.

        Convenience method; equivalent to calling fit(X) followed by
        predict(X).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to transform.

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

        Returns
        -------
        labels : ndarray of shape (n_samples,)
            Index of the cluster each sample belongs to.
        """
        return self.fit(
            X, sample_weight=sample_weight, session=session, run_kwargs=run_kwargs
        ).labels_

    def fit_transform(
        self, X, y=None, sample_weight=None, session=None, run_kwargs=None
    ):
        """Compute clustering and transform X to cluster-distance space.

        Equivalent to fit(X).transform(X), but more efficiently implemented.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to transform.

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

        Returns
        -------
        X_new : array of shape (n_samples, n_clusters)
            X transformed in the new space.
        """
        # Currently, this just skips a copy of the data if it is not in
        # np.array or CSR format already.
        # XXX This skips _check_test_data, which may change the dtype;
        # we should refactor the input validation.
        self.fit(X, sample_weight=sample_weight, session=session, run_kwargs=run_kwargs)
        return self._transform(X, session=session, run_kwargs=run_kwargs)

    def transform(self, X, session=None, run_kwargs=None):
        """Transform X to a cluster-distance space.

        In the new space, each dimension is the distance to the cluster
        centers.  Note that even if X is sparse, the array returned by
        `transform` will typically be dense.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to transform.

        Returns
        -------
        X_new : tensor of shape (n_samples, n_clusters)
            X transformed in the new space.
        """
        check_is_fitted(self)

        X = self._check_test_data(X)
        return self._transform(X, session=session, run_kwargs=run_kwargs)

    def _transform(self, X, session=None, run_kwargs=None):
        """guts of transform method; no input validation"""
        return euclidean_distances(X, self.cluster_centers_).execute(
            session=session, **(run_kwargs or dict())
        )

    def predict(self, X, sample_weight=None, session=None, run_kwargs=None):
        """Predict the closest cluster each sample in X belongs to.

        In the vector quantization literature, `cluster_centers_` is called
        the code book and each value returned by `predict` is the index of
        the closest code in the code book.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to predict.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

        Returns
        -------
        labels : tensor of shape (n_samples,)
            Index of the cluster each sample belongs to.
        """
        check_is_fitted(self)

        X = self._check_test_data(X)
        x_squared_norms = row_norms(X, squared=True)

        result = _labels_inertia(
            X,
            sample_weight,
            x_squared_norms,
            self.cluster_centers_,
            session=session,
            run_kwargs=run_kwargs,
        )[0]
        result.execute(session=session, *(run_kwargs or dict()))
        return result

    def score(self, X, y=None, sample_weight=None, session=None, run_kwargs=None):
        """Opposite of the value of X on the K-means objective.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data.

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

        Returns
        -------
        score : float
            Opposite of the value of X on the K-means objective.
        """
        check_is_fitted(self)

        X = self._check_test_data(X)
        x_squared_norms = row_norms(X, squared=True)

        result = -_labels_inertia(
            X,
            sample_weight,
            x_squared_norms,
            self.cluster_centers_,
            session=session,
            run_kwargs=run_kwargs,
        )[1]
        result.execute(session=session, **(run_kwargs or dict()))
        return result



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import weakref
from copy import deepcopy
from enum import Enum
from functools import partial
from typing import Any, List, Tuple, Dict, Type, Union

from ...serialization.serializables import (
    SerializableMeta,
    FieldTypes,
    BoolField,
    Int32Field,
    Float32Field,
    StringField,
    ListField,
    DictField,
    ReferenceField,
)
from ...serialization.core import Placeholder
from ...serialization.serializables import Serializable
from ...serialization.serializables.core import SerializableSerializer
from ...typing import OperandType
from ...utils import AttributeDict, classproperty, tokenize
from ..base import Base
from ..entity.core import Entity, EntityData
from ..entity.chunks import Chunk
from ..entity.tileables import Tileable
from ..entity.output_types import OutputType
from ..mode import enter_mode


class OperandMetaclass(SerializableMeta):
    def __new__(mcs, name: str, bases: Tuple[Type], properties: Dict):
        if "__call__" in properties:
            # if __call__ is specified for an operand,
            # make sure that entering user space
            properties["__call__"] = enter_mode(kernel=False)(properties["__call__"])

        return super().__new__(mcs, name, bases, properties)


class OperandStage(Enum):
    map = 0
    reduce = 1
    combine = 2
    agg = 3


class SchedulingHint(Serializable):
    # worker to execute, only work for chunk op,
    # if specified, the op should be executed on the specified worker
    # only work for those operand that has no input
    expect_worker = StringField("expect_worker", default=None)
    # will this operand be assigned a worker or not
    reassign_worker = BoolField("reassign_worker", default=False)
    # mark a op as fuseable
    fuseable = BoolField("fuseable", default=True)
    # True means control dependency, False means data dependency
    _pure_depends = ListField("pure_depends", FieldTypes.bool, default=None)
    # useful when setting chunk index as priority,
    # useful for those op like read_csv, the first chunk
    # need to be executed not later than the later ones,
    # because the range index of later chunk should be accumulated from
    # indexes of previous ones
    # `gpu` indicates that if the operand should be executed on the GPU.
    gpu = BoolField("gpu", default=None)
    priority = Int32Field("priority", default=None)

    @classproperty
    @functools.lru_cache(1)
    def all_hint_names(cls):
        return list(cls._FIELDS)

    def can_be_fused(self) -> bool:
        if not self.fuseable:
            return False
        if self.reassign_worker:
            return False
        if self._pure_depends and any(depend for depend in self._pure_depends):
            # control dependency exists
            return False
        return True


def _install_scheduling_hint_properties(cls: Type["Operand"]):
    def get_hint(name):
        def _get_val(operand: "Operand"):
            if operand.scheduling_hint:
                return getattr(operand.scheduling_hint, name)

        def _set_val(operand: "Operand", val: Any):
            if not operand.scheduling_hint:
                operand.scheduling_hint = SchedulingHint(**{name: val})
            else:
                setattr(operand.scheduling_hint, name, val)

        return property(_get_val, _set_val)

    for hint_name in SchedulingHint.all_hint_names:
        setattr(cls, hint_name, get_hint(hint_name))
    return cls


class OperatorLogicKeyGeneratorMixin:
    """
    This generator will generate an unique and deterministic key for operator compute logic. It should be same
    for different run if the compute logic doesn't change. This id will be used in subtask speculative
    execution and hbo scheduling and so on.
    """

    def get_logic_key(self):
        """The subclass may need to override this method to ensure unique and deterministic."""
        fields = self._get_logic_key_token_values()
        try:
            return tokenize(*fields)
        except Exception as e:  # pragma: no cover
            raise ValueError(
                f"Cannot generate logic key for operator {self} with fields {fields}"
            ) from e

    def _get_logic_key_token_values(self):
        token_values = [type(self).__module__, type(self).__name__]
        if self.stage is not None:
            token_values.append(self.stage.name)
        return token_values


class LogicKeyGenerator:
    def __init__(self):
        self.operator_id_to_logic_key = {}

    def get_logic_key(self, op: "Operand"):
        assert isinstance(op, Operand)
        logic_key = self.operator_id_to_logic_key.get(op.id)
        if logic_key is None:
            logic_key = self.operator_id_to_logic_key[op.id] = op.get_logic_key()
        return logic_key


@_install_scheduling_hint_properties
class Operand(Base, OperatorLogicKeyGeneratorMixin, metaclass=OperandMetaclass):
    """
    Operand base class. All operands should have a type, which can be Add, Subtract etc.
    `sparse` indicates that if the operand is applied on a sparse tensor/chunk.
    `device`, 0 means the CPU, otherwise means the GPU device.
    Operand can have inputs and outputs
    which should be the :class:`mars.tensor.core.TensorData`, :class:`mars.tensor.core.ChunkData` etc.
    """

    attr_tag = "attr"
    _init_update_key_ = False
    _output_type_ = None
    _no_copy_attrs_ = Base._no_copy_attrs_ | {"scheduling_hint"}
    _cache_primitive_serial = True

    sparse = BoolField("sparse", default=False)
    device = Int32Field("device", default=None)
    # will this operand create a view of input data or not
    create_view = BoolField("create_view", default=False)
    stage = ReferenceField("stage", OperandStage, default=None)
    memory_scale = Float32Field("memory_scale", default=None)
    tileable_op_key = StringField("tileable_op_key", default=None)
    extra_params = DictField("extra_params", key_type=FieldTypes.string)
    # scheduling hint
    scheduling_hint = ReferenceField("scheduling_hint", SchedulingHint, default=None)

    _inputs = ListField(
        "inputs", FieldTypes.reference(EntityData), default_factory=list
    )
    # outputs are weak-refs which are not pickle-able
    _outputs = ListField(
        "outputs", default=None, on_serialize=lambda outputs: [o() for o in outputs]
    )
    _output_types = ListField(
        "output_type", FieldTypes.reference(OutputType), default=None
    )

    def __init__(self: OperandType, *args, **kwargs):
        self._parse_kwargs(kwargs)
        super().__init__(*args, **kwargs)

    @classmethod
    def _parse_kwargs(cls, kwargs: Dict[str, Any]):
        extra_params = kwargs.pop("extra_params", {})
        kwargs["extra_params"] = extras = AttributeDict(extra_params)
        kwargs["scheduling_hint"] = scheduling_hint = kwargs.get(
            "scheduling_hint", SchedulingHint()
        )
        for k in set(kwargs):
            if k in cls._FIELDS:
                continue
            elif k in SchedulingHint.all_hint_names:
                setattr(scheduling_hint, k, kwargs.pop(k))
            else:
                extras[k] = kwargs.pop(k)

    def __repr__(self):
        if self.stage is None:
            return f"{type(self).__name__} <key={self.key}>"
        else:
            return f"{type(self).__name__} <key={self.key}, stage={self.stage.name}>"

    @classmethod
    def _get_entity_data(cls, entity):
        if isinstance(entity, Entity):
            return entity.data
        return entity

    @classmethod
    def _get_inputs_data(cls, inputs):
        return [cls._get_entity_data(inp) for inp in inputs]

    def _set_inputs(self, inputs):
        if inputs is not None:
            inputs = self._get_inputs_data(inputs)
        if hasattr(self, "check_inputs"):
            self.check_inputs(inputs)
        setattr(self, "_inputs", inputs)

    @property
    def inputs(self) -> List[Union[Chunk, Tileable]]:
        inputs = self._inputs
        if inputs is None:
            inputs = self._inputs = []
        return inputs

    @inputs.setter
    def inputs(self, vals):
        self._set_inputs(vals)

    @property
    def output_limit(self) -> int:
        return 1

    @property
    def pure_depends(self):
        val = self._pure_depends  # pylint: disable=access-member-before-definition
        if not val:
            val = self._pure_depends = [False] * len(self.inputs or ())
        return val

    @property
    def output_types(self):
        return self._output_types

    @output_types.setter
    def output_types(self, value):
        self._output_types = value

    def _attach_outputs(self, *outputs):
        self._outputs = [
            weakref.ref(self._get_entity_data(o)) if o is not None else o
            for o in outputs
        ]

        if len(self._outputs) > self.output_limit:
            raise ValueError("Outputs' size exceeds limitation")

    @property
    def outputs(self) -> List[Union[Chunk, Tileable]]:
        outputs = self._outputs
        if outputs:
            return [ref() for ref in outputs]

    @outputs.setter
    def outputs(self, outputs):
        self._attach_outputs(*outputs)

    def is_sparse(self) -> bool:
        return self.sparse

    issparse = is_sparse

    def is_gpu(self) -> bool:
        return self.gpu

    @property
    def retryable(self) -> bool:
        return True

    def get_dependent_data_keys(self):
        return [dep.key for dep in self.inputs or ()]

    def _get_output_type(self, output_idx):
        if self.output_types:
            try:
                return self.output_types[output_idx]
            except IndexError:
                return self.output_types[0]
        else:
            return self._output_type_

    def copy(self: OperandType) -> OperandType:
        new_op = super().copy()
        new_op.outputs = []
        # copy scheduling_hint
        new_op.scheduling_hint = self.scheduling_hint.copy()
        extra_params = self.extra_params
        if extra_params:
            new_op.extra_params = deepcopy(extra_params)
        return new_op

    def on_output_modify(self, new_output):
        # when `create_view` is True, if the output is modified,
        # the modification should be set back to the input.
        # This function is for this sort of usage.
        # Remember, if `create_view` is False, this function should take no effect.
        raise NotImplementedError

    def on_input_modify(self, new_input):
        # when `create_view` is True, if the input is modified,
        # this function could be used to respond the modification.
        # Remember, if `create_view` is False, this function should take no effect.
        raise NotImplementedError


class OperandSerializer(SerializableSerializer):
    def serial(self, obj: Serializable, context: Dict):
        res = super().serial(obj, context)
        return res

    def deserial(self, serialized: Tuple, context: Dict, subs: List) -> Operand:
        # convert outputs back to weak-refs
        operand: Operand = super().deserial(serialized, context, subs)
        for i, out in enumerate(operand._outputs):

            def cb(o, index):
                outputs = operand._outputs
                outputs[index] = weakref.ref(o)

                if len(outputs) > 1 and all(
                    not isinstance(o, Placeholder) for o in outputs
                ):
                    # all replaced
                    # add siblings for multiple outputs
                    outputs = operand.outputs
                    for j in range(len(outputs)):
                        outputs[j]._siblings = outputs[:j] + outputs[j + 1 :]

            if isinstance(out, Placeholder):
                out.callbacks.append(partial(cb, index=i))
            else:
                cb(out, i)
        return operand


OperandSerializer.register(Operand)


class VirtualOperand(Operand):
    def get_dependent_data_keys(self):
        return []


class HasInput(Operand):
    __slots__ = ()

    @property
    def input(self):
        return self._input

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers
from abc import ABCMeta, abstractmethod

from numpy.linalg import LinAlgError
from sklearn.utils.validation import check_is_fitted, _deprecate_positional_args
from sklearn.base import MultiOutputMixin

from ... import execute
from ... import tensor as mt
from ...tensor.datasource import tensor as astensor
from ..base import BaseEstimator, RegressorMixin, ClassifierMixin
from ..preprocessing import normalize as f_normalize
from ..utils.validation import _check_sample_weight, check_array, FLOAT_DTYPES


def _preprocess_data(
    X,
    y,
    fit_intercept,
    normalize=False,
    copy=True,
    sample_weight=None,
    return_mean=False,
    check_input=True,
):
    """Center and scale data.

    Centers data to have mean zero along axis 0. If fit_intercept=False or if
    the X is a sparse matrix, no centering is done, but normalization can still
    be applied. The function returns the statistics necessary to reconstruct
    the input data, which are X_offset, y_offset, X_scale, such that the output

        X = (X - X_offset) / X_scale

    X_scale is the L2 norm of X - X_offset. If sample_weight is not None,
    then the weighted mean of X and y is zero, and not the mean itself. If
    return_mean=True, the mean, eventually weighted, is returned, independently
    of whether X was centered (option used for optimization with sparse data in
    coordinate_descend).

    This is here because nearly all linear models will want their data to be
    centered. This function also systematically makes y consistent with X.dtype
    """
    if isinstance(sample_weight, numbers.Number):
        sample_weight = None
    if sample_weight is not None:
        sample_weight = astensor(sample_weight)

    X = astensor(X)
    y = astensor(y, dtype=X.dtype)

    if check_input:
        X = check_array(X, copy=copy, accept_sparse=["csr", "csc"], dtype=FLOAT_DTYPES)
    elif copy:
        if X.issparse():
            X = X.copy()
        else:
            X = X.copy(order="K")

    if fit_intercept:
        if X.issparse():
            raise NotImplementedError("Does not support sparse input!")
        else:
            X_offset = mt.average(X, axis=0, weights=sample_weight)
            X = X - X_offset
            if normalize:
                X, X_scale = f_normalize(X, axis=0, copy=False, return_norm=True)
            else:
                X_scale = mt.ones(X.shape[1], dtype=X.dtype)
        y_offset = mt.average(y, axis=0, weights=sample_weight)
        y = y - y_offset
    else:
        if X.issparse():
            raise NotImplementedError("Does not support sparse input!")
        X_offset = mt.zeros(X.shape[1], dtype=X.dtype)
        X_scale = mt.ones(X.shape[1], dtype=X.dtype)
        if y.ndim == 1:
            y_offset = X.dtype.type(0)
        else:
            y_offset = mt.zeros(y.shape[1], dtype=X.dtype)

    return X, y, X_offset, y_offset, X_scale


def _rescale_data(X, y, sample_weight):
    """Rescale data sample-wise by square root of sample_weight.

    For many linear models, this enables easy support for sample_weight.

    Returns
    -------
    X_rescaled : {array-like, sparse matrix}

    y_rescaled : {array-like, sparse matrix}
    """
    n_samples = X.shape[0]
    sample_weight = mt.asarray(sample_weight)
    if sample_weight.ndim == 0:
        sample_weight = mt.full(n_samples, sample_weight, dtype=sample_weight.dtype)
    sample_weight = mt.sqrt(sample_weight)
    sw_matrix = mt.diag(sample_weight, sparse=True)
    X = mt.dot(sw_matrix, X)
    y = mt.dot(sw_matrix, y)
    return X, y


class LinearModel(BaseEstimator, metaclass=ABCMeta):
    """Base class for Linear Models"""

    @abstractmethod
    def fit(self, X, y, sample_weight=None):
        """Fit model."""

    def _decision_function(self, X):
        check_is_fitted(self)

        X = self._validate_data(
            X, y="no_validation", accept_sparse=["csr", "csc", "coo"], reset=False
        )
        return mt.dot(X, self.coef_.T) + self.intercept_

    def predict(self, X):
        """
        Predict using the linear model.

        Parameters
        ----------
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape (n_samples,)
            Returns predicted values.
        """
        return self._decision_function(X)

    _preprocess_data = staticmethod(_preprocess_data)

    def _set_intercept(self, X_offset, y_offset, X_scale):
        """Set the intercept_"""
        if self.fit_intercept:
            self.coef_ = self.coef_ / X_scale
            self.intercept_ = y_offset - mt.dot(X_offset, self.coef_.T)
            execute(self.coef_, self.intercept_)
        else:
            self.intercept_ = mt.tensor(0.0)
            self.intercept_.execute()

    def _more_tags(self):  # noqa: R0201  # pylint: disable=no-self-use
        return {"requires_y": True}


class LinearRegression(MultiOutputMixin, RegressorMixin, LinearModel):
    """
    Ordinary least squares Linear Regression.

    LinearRegression fits a linear model with coefficients w = (w1, ..., wp)
    to minimize the residual sum of squares between the observed targets in
    the dataset, and the targets predicted by the linear approximation.

    Parameters
    ----------
    fit_intercept : bool, default=True
        Whether to calculate the intercept for this model. If set
        to False, no intercept will be used in calculations
        (i.e. data is expected to be centered).

    normalize : bool, default=False
        This parameter is ignored when ``fit_intercept`` is set to False.
        If True, the regressors X will be normalized before regression by
        subtracting the mean and dividing by the l2-norm.
        If you wish to standardize, please use
        :class:`~sklearn.preprocessing.StandardScaler` before calling ``fit``
        on an estimator with ``normalize=False``.

    copy_X : bool, default=True
        If True, X will be copied; else, it may be overwritten.

    positive : bool, default=False
        When set to ``True``, forces the coefficients to be positive. This
        option is only supported for dense arrays.

    Attributes
    ----------
    coef_ : array of shape (n_features, ) or (n_targets, n_features)
        Estimated coefficients for the linear regression problem.
        If multiple targets are passed during the fit (y 2D), this
        is a 2D array of shape (n_targets, n_features), while if only
        one target is passed, this is a 1D array of length n_features.

    rank_ : int
        Rank of matrix `X`. Only available when `X` is dense.

    singular_ : array of shape (min(X, y),)
        Singular values of `X`. Only available when `X` is dense.

    intercept_ : float or array of shape (n_targets,)
        Independent term in the linear model. Set to 0.0 if
        `fit_intercept = False`.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    See Also
    --------
    Ridge : Ridge regression addresses some of the
        problems of Ordinary Least Squares by imposing a penalty on the
        size of the coefficients with l2 regularization.
    Lasso : The Lasso is a linear model that estimates
        sparse coefficients with l1 regularization.
    ElasticNet : Elastic-Net is a linear regression
        model trained with both l1 and l2 -norm regularization of the
        coefficients.
    """

    @_deprecate_positional_args
    def __init__(
        self,
        *,
        fit_intercept=True,
        normalize=False,
        copy_X=True,
        positive=False,
    ):
        self.fit_intercept = fit_intercept
        self.normalize = normalize
        self.copy_X = copy_X
        self.positive = positive

    def fit(self, X, y, sample_weight=None):
        """
        Fit linear model.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            Target values. Will be cast to X's dtype if necessary.

        sample_weight : array-like of shape (n_samples,), default=None
            Individual weights for each sample.

        Returns
        -------
        self : object
            Fitted Estimator.
        """
        accept_sparse = False if self.positive else ["csr", "csc", "coo"]

        X, y = self._validate_data(
            X, y, accept_sparse=accept_sparse, y_numeric=True, multi_output=True
        )

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, dtype=X.dtype)

        X, y, X_offset, y_offset, X_scale = self._preprocess_data(
            X,
            y,
            fit_intercept=self.fit_intercept,
            normalize=self.normalize,
            copy=self.copy_X,
            sample_weight=sample_weight,
            return_mean=True,
        )

        if sample_weight is not None:
            # Sample weight can be implemented via a simple rescaling.
            X, y = _rescale_data(X, y, sample_weight)

        if self.positive:
            # TODO: implement optimize.nnls first
            raise NotImplementedError("Does not support positive coefficients!")
        elif X.issparse():
            # TODO: implement sparse.linalg.lsqr first
            raise NotImplementedError("Does not support sparse input!")
        else:
            try:
                # In numpy:
                #   Mat mul does NOT always satisfy associative law
                # Tyipical mistake:
                #   (mt.linalg.inv(X.T @ X) @ (X.T @ y)).T
                self.coef_ = (mt.linalg.inv(X.T @ X) @ X.T @ y).T
                self.coef_.execute()
            except LinAlgError:
                # TODO: implement linalg.lstsq first
                raise NotImplementedError("Does not support sigular matrix!")

        if y.ndim == 1:
            self.coef_ = mt.ravel(self.coef_)
            self.coef_.execute()
        self._set_intercept(X_offset, y_offset, X_scale)
        return self


class LinearClassifierMixin(ClassifierMixin):
    """Mixin for linear classifiers.

    Handles prediction for sparse and dense X.
    """

    def decision_function(self, X):
        """
        Predict confidence scores for samples.

        The confidence score for a sample is proportional to the signed
        distance of that sample to the hyperplane.

        Parameters
        ----------
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)
            Confidence scores per (sample, class) combination. In the binary
            case, confidence score for self.classes_[1] where >0 means this
            class would be predicted.
        """
        check_is_fitted(self)

        X = check_array(X, accept_sparse="csr")

        n_features = self.coef_.shape[1]
        if X.shape[1] != n_features:
            raise ValueError(
                "X has %d features per sample; expecting %d" % (X.shape[1], n_features)
            )

        scores = mt.dot(X, self.coef_.T) + self.intercept_
        return scores

    def predict(self, X):
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape [n_samples]
            Predicted class label per sample.
        """
        scores = self.decision_function(X)
        indices = scores.argmax(axis=1)
        return self.classes_[indices].execute()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import IOBase

from ....utils import lazy_import
from .common import parse_osspath, oss_stat

oss2 = lazy_import("oss2", placeholder=True)


class OSSIOBase(IOBase):
    def __init__(self, path, mode):
        self._path = path
        (
            self._bucket_name,
            self._key_name,
            self._access_key_id,
            self._access_key_secret,
            self._end_point,
        ) = parse_osspath(self._path)
        self._bucket = self._get_bucket()
        self._current_pos = 0
        self._size = None
        self._buffer = b""
        self._buffer_size = 1 * 1024
        self._mode = mode

    @property
    def mode(self):
        return self._mode

    def fileno(self) -> int:
        raise AttributeError

    def _get_bucket(self):
        return oss2.Bucket(
            auth=oss2.Auth(
                access_key_id=self._access_key_id,
                access_key_secret=self._access_key_secret,
            ),
            endpoint=self._end_point,
            bucket_name=self._bucket_name,
        )

    def _get_size(self):
        if self._size is None:
            self._size = int(oss_stat(self._path)["size"])
        return self._size

    def seek(self, pos, whence=0):
        if whence == 0:
            if pos < 0:
                raise OSError("Invalid argument")
            self._current_pos = pos
        elif whence == 2:
            self._current_pos = self._get_size() + pos
        elif whence == 1:
            check_pos = self._current_pos + pos
            if check_pos < 0:
                raise OSError("Invalid argument")
            else:
                self._current_pos = self._current_pos + pos
        else:
            raise ValueError('Parameter "whence" should be 0 or 1 or 2')
        if pos > 0 and self._current_pos > self._get_size() - 1:
            self._current_pos = self._get_size()
        return self._current_pos

    def seekable(self):
        return True

    def read(self, size=-1):
        """
        Read and return up to size bytes, where size is an int.

        If the argument is omitted, None, or negative, reads and
        returns all data until EOF.

        If the argument is positive, multiple raw reads may be issued to satisfy
        the byte count (unless EOF is reached first).

        Returns an empty bytes array on EOF.
        """
        if self._current_pos == self._get_size() or size == 0:
            return b""
        elif size < 0:
            obj = self._bucket.get_object(
                self._key_name, byte_range=(self._current_pos, None)
            )
            self._current_pos = self._get_size()
        else:
            obj = self._bucket.get_object(
                self._key_name,
                byte_range=(self._current_pos, self._current_pos + size - 1),
            )
            self._current_pos = self._current_pos + size
        content = obj.read()
        return content

    def readline(self, size=-1):
        # For backwards compatibility, a (slowish) readline().
        def nreadahead():
            # Read to the beginning of the next line
            read_to = min(
                self._get_size() - 1, self._current_pos + self._buffer_size - 1
            )
            buffer = self._bucket.get_object(
                self._key_name, byte_range=(self._current_pos, read_to)
            ).read()
            if not buffer:
                return 1
            n = (buffer.find(b"\n") + 1) or len(buffer)
            if size >= 0:
                n = min(n, size)
            return n

        if size is None:
            size = -1
        else:
            try:
                size_index = size.__index__
            except AttributeError:
                raise TypeError(f"{size!r} is not an integer")
            else:
                size = size_index()
        res = bytearray()
        while size < 0 or len(res) < size:
            b = self.read(nreadahead())
            if not b:
                break
            res += b
            if res.endswith(b"\n"):
                break
        return bytes(res)

    def readable(self):
        return True

    def writable(self):
        return False

    def close(self):
        # already closed by oss
        pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, OutputType, recursive_tile
from ...serialization.serializables import (
    FieldTypes,
    ListField,
    StringField,
    BoolField,
    AnyField,
)
from ...utils import lazy_import, has_unknown_shape
from ..operands import DataFrameOperand, DataFrameOperandMixin, SERIES_TYPE
from ..utils import (
    parse_index,
    build_empty_df,
    build_empty_series,
    standardize_range_index,
    validate_axis,
)

cudf = lazy_import("cudf")


class DataFrameConcat(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.CONCATENATE

    _axis = AnyField("axis")
    _join = StringField("join")
    _ignore_index = BoolField("ignore_index")
    _keys = ListField("keys")
    _levels = ListField("levels")
    _names = ListField("names")
    _verify_integrity = BoolField("verify_integrity")
    _sort = BoolField("sort")
    _copy = BoolField("copy")

    def __init__(
        self,
        axis=None,
        join=None,
        ignore_index=None,
        keys=None,
        levels=None,
        names=None,
        verify_integrity=None,
        sort=None,
        copy=None,
        sparse=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _axis=axis,
            _join=join,
            _ignore_index=ignore_index,
            _keys=keys,
            _levels=levels,
            _names=names,
            _verify_integrity=verify_integrity,
            _sort=sort,
            _copy=copy,
            _output_types=output_types,
            sparse=sparse,
            **kw
        )

    @property
    def axis(self):
        return self._axis

    @property
    def join(self):
        return self._join

    @property
    def ignore_index(self):
        return self._ignore_index

    @property
    def keys(self):
        return self._keys

    @property
    def level(self):
        return self._levels

    @property
    def name(self):
        return self._names

    @property
    def verify_integrity(self):
        return self._verify_integrity

    @property
    def sort(self):
        return self._sort

    @property
    def copy_(self):
        return self._copy

    @classmethod
    def _tile_dataframe(cls, op):
        from ..indexing.iloc import DataFrameIlocGetItem

        out_df = op.outputs[0]
        inputs = op.inputs
        axis = op.axis

        if not all(
            inputs[i].nsplits[1 - axis] == inputs[i + 1].nsplits[1 - axis]
            for i in range(len(inputs) - 1)
        ):
            # need rechunk
            if has_unknown_shape(*inputs):
                yield
            normalized_nsplits = {1 - axis: inputs[0].nsplits[1 - axis]}
            new_inputs = []
            for inp in inputs:
                new_inputs.append(
                    (yield from recursive_tile(inp.rechunk(normalized_nsplits)))
                )
            inputs = new_inputs

        out_chunks = []
        nsplits = []
        cum_index = 0
        for df in inputs:
            for c in df.chunks:
                if op.axis == 0:
                    index = (c.index[0] + cum_index, c.index[1])
                else:
                    index = (c.index[0], c.index[1] + cum_index)

                iloc_op = DataFrameIlocGetItem(indexes=[slice(None)] * 2)
                out_chunks.append(
                    iloc_op.new_chunk(
                        [c],
                        shape=c.shape,
                        index=index,
                        dtypes=c.dtypes,
                        index_value=c.index_value,
                        columns_value=c.columns_value,
                    )
                )
            nsplits.extend(df.nsplits[op.axis])
            cum_index += len(df.nsplits[op.axis])
        out_nsplits = (
            (tuple(nsplits), inputs[0].nsplits[1])
            if op.axis == 0
            else (inputs[0].nsplits[0], tuple(nsplits))
        )

        if op.ignore_index:
            yield out_chunks
            out_chunks = standardize_range_index(out_chunks)

        new_op = op.copy()
        return new_op.new_dataframes(
            op.inputs,
            out_df.shape,
            nsplits=out_nsplits,
            chunks=out_chunks,
            dtypes=out_df.dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
        )

    @classmethod
    def _tile_series(cls, op):
        from ..datasource.from_tensor import DataFrameFromTensor
        from ..indexing.iloc import SeriesIlocGetItem, DataFrameIlocGetItem

        out = op.outputs[0]
        inputs = op.inputs
        out_chunks = []

        if op.axis == 1:
            if has_unknown_shape(*inputs):
                yield
            new_inputs = []
            for inp in inputs:
                new_inputs.append(
                    (yield from recursive_tile(inp.rechunk(op.inputs[0].nsplits)))
                )
            inputs = new_inputs

        cum_index = 0
        offset = 0
        nsplits = []
        for series in inputs:
            for c in series.chunks:
                if op.axis == 0:
                    index = (c.index[0] + cum_index,)
                    shape = c.shape
                    iloc_op = SeriesIlocGetItem(indexes=(slice(None),))
                    out_chunks.append(
                        iloc_op.new_chunk(
                            [c],
                            shape=shape,
                            index=index,
                            index_value=c.index_value,
                            dtype=c.dtype,
                            name=c.name,
                        )
                    )
                else:
                    index = (c.index[0], cum_index)
                    shape = (c.shape[0], 1)
                    to_frame_op = DataFrameFromTensor(
                        input=c,
                        index=None,
                        columns=None,
                    )
                    if c.name:
                        dtypes = pd.Series([c.dtype], index=[c.name])
                    else:
                        dtypes = pd.Series(
                            [c.dtype], index=pd.RangeIndex(offset, offset + 1)
                        )
                    df_chunk = to_frame_op.new_chunk(
                        [c],
                        shape=shape,
                        index=index,
                        index_value=c.index_value,
                        columns_value=parse_index(dtypes.index, store_data=True),
                        dtypes=dtypes,
                    )
                    iloc_op = DataFrameIlocGetItem(indexes=[slice(None)] * 2)
                    out_chunks.append(
                        iloc_op.new_chunk(
                            [df_chunk],
                            shape=df_chunk.shape,
                            index=index,
                            dtypes=df_chunk.dtypes,
                            index_value=df_chunk.index_value,
                            columns_value=df_chunk.columns_value,
                        )
                    )

            if op.axis == 0:
                nsplits.extend(series.nsplits[0])
                cum_index += len(series.nsplits[op.axis])
            else:
                nsplits.append(1)
                cum_index += 1
                offset += 1

        if op.ignore_index:
            yield out_chunks
            out_chunks = standardize_range_index(out_chunks)

        new_op = op.copy()
        if op.axis == 0:
            nsplits = (tuple(nsplits),)
            return new_op.new_seriess(
                op.inputs,
                out.shape,
                nsplits=nsplits,
                chunks=out_chunks,
                dtype=out.dtype,
                index_value=out.index_value,
                name=out.name,
            )
        else:
            nsplits = (inputs[0].nsplits[0], tuple(nsplits))
            return new_op.new_dataframes(
                op.inputs,
                out.shape,
                nsplits=nsplits,
                chunks=out_chunks,
                dtypes=out.dtypes,
                index_value=out.index_value,
                columns_value=out.columns_value,
            )

    @classmethod
    def tile(cls, op):
        if isinstance(op.inputs[0], SERIES_TYPE):
            return (yield from cls._tile_series(op))
        else:
            return (yield from cls._tile_dataframe(op))

    @classmethod
    def execute(cls, ctx, op):
        def _base_concat(chunk, inputs):
            # auto generated concat when executing a DataFrame, Series or Index
            if chunk.op.output_types[0] == OutputType.dataframe:
                return _auto_concat_dataframe_chunks(chunk, inputs)
            elif chunk.op.output_types[0] == OutputType.series:
                return _auto_concat_series_chunks(chunk, inputs)
            elif chunk.op.output_types[0] == OutputType.index:
                return _auto_concat_index_chunks(chunk, inputs)
            elif chunk.op.output_types[0] == OutputType.categorical:
                return _auto_concat_categorical_chunks(chunk, inputs)
            else:  # pragma: no cover
                raise TypeError(
                    "Only DataFrameChunk, SeriesChunk, IndexChunk, "
                    "and CategoricalChunk can be automatically concatenated"
                )

        def _auto_concat_dataframe_chunks(chunk, inputs):
            xdf = (
                pd
                if isinstance(inputs[0], (pd.DataFrame, pd.Series)) or cudf is None
                else cudf
            )

            if chunk.op.axis is not None:
                return xdf.concat(inputs, axis=op.axis)

            # auto generated concat when executing a DataFrame
            if len(inputs) == 1:
                ret = inputs[0]
            else:
                n_rows = len(set(inp.index[0] for inp in chunk.inputs))
                n_cols = int(len(inputs) // n_rows)
                assert n_rows * n_cols == len(inputs)

                concats = []
                for i in range(n_rows):
                    if n_cols == 1:
                        concats.append(inputs[i])
                    else:
                        concat = xdf.concat(
                            [inputs[i * n_cols + j] for j in range(n_cols)], axis=1
                        )
                        concats.append(concat)

                if xdf is pd:
                    # The `sort=False` is to suppress a `FutureWarning` of pandas,
                    # when the index or column of chunks to concatenate is not aligned,
                    # which may happens for certain ops.
                    #
                    # See also Note [Columns of Left Join] in test_merge_execution.py.
                    ret = xdf.concat(concats, sort=False)
                else:
                    ret = xdf.concat(concats)
                    # cuDF will lost index name when concat two seriess.
                    ret.index.name = concats[0].index.name

            return ret

        def _auto_concat_series_chunks(chunk, inputs):
            # auto generated concat when executing a Series
            if len(inputs) == 1:
                concat = inputs[0]
            else:
                xdf = pd if isinstance(inputs[0], pd.Series) or cudf is None else cudf
                if chunk.op.axis is not None:
                    concat = xdf.concat(inputs, axis=chunk.op.axis)
                else:
                    concat = xdf.concat(inputs)
            return concat

        def _auto_concat_index_chunks(chunk, inputs):
            if len(inputs) == 1:
                xdf = pd if isinstance(inputs[0], pd.Index) or cudf is None else cudf
                concat_df = xdf.DataFrame(index=inputs[0])
            else:
                xdf = pd if isinstance(inputs[0], pd.Index) or cudf is None else cudf
                empty_dfs = [xdf.DataFrame(index=inp) for inp in inputs]
                concat_df = xdf.concat(empty_dfs, axis=0)
            return concat_df.index

        def _auto_concat_categorical_chunks(_, inputs):
            if len(inputs) == 1:  # pragma: no cover
                return inputs[0]
            else:
                # convert categorical into array
                arrays = [np.asarray(inp) for inp in inputs]
                array = np.concatenate(arrays)
                return pd.Categorical(
                    array, categories=inputs[0].categories, ordered=inputs[0].ordered
                )

        chunk = op.outputs[0]
        inputs = [ctx[input.key] for input in op.inputs]

        if isinstance(inputs[0], tuple):
            ctx[chunk.key] = tuple(
                _base_concat(chunk, [input[i] for input in inputs])
                for i in range(len(inputs[0]))
            )
        else:
            ctx[chunk.key] = _base_concat(chunk, inputs)

    @classmethod
    def _concat_index(cls, prev_index: pd.Index, cur_index: pd.Index):
        if isinstance(prev_index, pd.RangeIndex) and isinstance(
            cur_index, pd.RangeIndex
        ):
            # handle RangeIndex that append may generate huge amount of data
            # e.g. pd.RangeIndex(10_000) and pd.RangeIndex(10_000)
            # will generate a Int64Index full of data
            # for details see GH#1647
            prev_stop = prev_index.start + prev_index.size * prev_index.step
            cur_start = cur_index.start
            if prev_stop == cur_start and prev_index.step == cur_index.step:
                # continuous RangeIndex, still return RangeIndex
                return prev_index.append(cur_index)
            else:
                # otherwise, return an empty index
                return pd.Index([], dtype=prev_index.dtype)
        elif isinstance(prev_index, pd.RangeIndex):
            return pd.Index([], prev_index.dtype).append(cur_index)
        elif isinstance(cur_index, pd.RangeIndex):
            return prev_index.append(pd.Index([], cur_index.dtype))
        return prev_index.append(cur_index)

    def _call_series(self, objs):
        if self.axis == 0:
            row_length = 0
            index = None
            for series in objs:
                if index is None:
                    index = series.index_value.to_pandas()
                else:
                    index = self._concat_index(index, series.index_value.to_pandas())
                row_length += series.shape[0]
            if self.ignore_index:  # pragma: no cover
                index_value = parse_index(pd.RangeIndex(row_length))
            else:
                index_value = parse_index(index, objs)
            return self.new_series(
                objs,
                shape=(row_length,),
                dtype=objs[0].dtype,
                index_value=index_value,
                name=objs[0].name,
            )
        else:
            col_length = 0
            columns = []
            dtypes = dict()
            undefined_name = 0
            for series in objs:
                if series.name is None:
                    dtypes[undefined_name] = series.dtype
                    undefined_name += 1
                    columns.append(undefined_name)
                else:
                    dtypes[series.name] = series.dtype
                    columns.append(series.name)
                col_length += 1
            if self.ignore_index or undefined_name == len(objs):
                columns_value = parse_index(pd.RangeIndex(col_length))
            else:
                columns_value = parse_index(pd.Index(columns), store_data=True)

            shape = (objs[0].shape[0], col_length)
            return self.new_dataframe(
                objs,
                shape=shape,
                dtypes=pd.Series(dtypes),
                index_value=objs[0].index_value,
                columns_value=columns_value,
            )

    def _call_dataframes(self, objs):
        if self.axis == 0:
            row_length = 0
            index = None
            empty_dfs = []
            for df in objs:
                if index is None:
                    index = df.index_value.to_pandas()
                else:
                    index = self._concat_index(index, df.index_value.to_pandas())
                row_length += df.shape[0]
                if df.ndim == 2:
                    empty_dfs.append(build_empty_df(df.dtypes))
                else:
                    empty_dfs.append(build_empty_series(df.dtype, name=df.name))

            emtpy_result = pd.concat(empty_dfs, join=self.join, sort=True)
            shape = (row_length, emtpy_result.shape[1])
            columns_value = parse_index(emtpy_result.columns, store_data=True)

            if self.join == "inner":
                objs = [o[list(emtpy_result.columns)] for o in objs]

            if self.ignore_index:  # pragma: no cover
                index_value = parse_index(pd.RangeIndex(row_length))
            else:
                index_value = parse_index(index, objs)

            new_objs = []
            for obj in objs:
                if obj.ndim != 2:
                    # series
                    new_obj = obj.to_frame().reindex(columns=emtpy_result.dtypes.index)
                else:
                    # dataframe
                    if list(obj.dtypes.index) != list(emtpy_result.dtypes.index):
                        new_obj = obj.reindex(columns=emtpy_result.dtypes.index)
                    else:
                        new_obj = obj
                new_objs.append(new_obj)

            return self.new_dataframe(
                new_objs,
                shape=shape,
                dtypes=emtpy_result.dtypes,
                index_value=index_value,
                columns_value=columns_value,
            )
        else:
            col_length = 0
            empty_dfs = []
            for df in objs:
                if df.ndim == 2:
                    # DataFrame
                    col_length += df.shape[1]
                    empty_dfs.append(build_empty_df(df.dtypes))
                else:
                    # Series
                    col_length += 1
                    empty_dfs.append(build_empty_series(df.dtype, name=df.name))

            emtpy_result = pd.concat(empty_dfs, join=self.join, axis=1, sort=True)
            if self.ignore_index:
                columns_value = parse_index(pd.RangeIndex(col_length))
            else:
                columns_value = parse_index(
                    pd.Index(emtpy_result.columns), store_data=True
                )

            if self.ignore_index or len({o.index_value.key for o in objs}) == 1:
                new_objs = [obj if obj.ndim == 2 else obj.to_frame() for obj in objs]
            else:  # pragma: no cover
                raise NotImplementedError(
                    "Does not support concat dataframes which has different index"
                )

            shape = (objs[0].shape[0], col_length)
            return self.new_dataframe(
                new_objs,
                shape=shape,
                dtypes=emtpy_result.dtypes,
                index_value=objs[0].index_value,
                columns_value=columns_value,
            )

    def __call__(self, objs):
        if all(isinstance(obj, SERIES_TYPE) for obj in objs):
            self.output_types = [OutputType.series]
            return self._call_series(objs)
        else:
            self.output_types = [OutputType.dataframe]
            return self._call_dataframes(objs)


class GroupByConcat(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.GROUPBY_CONCAT

    _groups = ListField("groups", FieldTypes.key)
    _groupby_params = AnyField("groupby_params")

    def __init__(self, groups=None, groupby_params=None, output_types=None, **kw):
        super().__init__(
            _groups=groups,
            _groupby_params=groupby_params,
            _output_types=output_types,
            **kw
        )

    @property
    def groups(self):
        return self._groups

    @property
    def groupby_params(self):
        return self._groupby_params

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)

        new_groups = []
        for _ in self._groups:
            new_groups.append(next(inputs_iter))
        self._groups = new_groups

        if isinstance(self._groupby_params["by"], list):
            by = []
            for v in self._groupby_params["by"]:
                if isinstance(v, ENTITY_TYPE):
                    by.append(next(inputs_iter))
                else:
                    by.append(v)
            self._groupby_params["by"] = by

    @classmethod
    def execute(cls, ctx, op):
        input_data = [ctx[input.key] for input in op.groups]
        obj = pd.concat([d.obj for d in input_data])

        params = op.groupby_params.copy()
        if isinstance(params["by"], list):
            by = []
            for v in params["by"]:
                if isinstance(v, ENTITY_TYPE):
                    by.append(ctx[v.key])
                else:
                    by.append(v)
            params["by"] = by
        selection = params.pop("selection", None)

        result = obj.groupby(**params)
        if selection:
            result = result[selection]

        ctx[op.outputs[0].key] = result


def concat(
    objs,
    axis=0,
    join="outer",
    ignore_index=False,
    keys=None,
    levels=None,
    names=None,
    verify_integrity=False,
    sort=False,
    copy=True,
):
    if not isinstance(objs, (list, tuple)):  # pragma: no cover
        raise TypeError(
            "first argument must be an iterable of dataframe or series objects"
        )
    axis = validate_axis(axis)
    if isinstance(objs, dict):  # pragma: no cover
        keys = objs.keys()
        objs = objs.values()
    if axis == 1 and join == "inner":  # pragma: no cover
        raise NotImplementedError("inner join is not support when specify `axis=1`")
    if verify_integrity or sort or keys:  # pragma: no cover
        raise NotImplementedError(
            "verify_integrity, sort, keys arguments are not supported now"
        )
    op = DataFrameConcat(
        axis=axis,
        join=join,
        ignore_index=ignore_index,
        keys=keys,
        levels=levels,
        names=names,
        verify_integrity=verify_integrity,
        sort=sort,
        copy=copy,
    )

    return op(objs)



# distutils: language = c++
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import os
import pickle
import pkgutil
import time
import types
import uuid
from datetime import date, datetime, timedelta, tzinfo
from enum import Enum
from functools import lru_cache, partial
from random import getrandbits
from weakref import WeakSet

import numpy as np
import pandas as pd
import cloudpickle
cimport cython
from libc.stdint cimport uint_fast64_t
from libc.stdlib cimport malloc, free
from .lib.cython.libcpp cimport mt19937_64
try:
    from pandas.tseries.offsets import Tick as PDTick
except ImportError:
    PDTick = None

from .lib.mmh3 import hash as mmh_hash, hash_bytes as mmh_hash_bytes, \
    hash_from_buffer as mmh3_hash_from_buffer

cdef bint _has_cupy = bool(pkgutil.find_loader('cupy'))
cdef bint _has_cudf = bool(pkgutil.find_loader('cudf'))
cdef bint _has_sqlalchemy = bool(pkgutil.find_loader('sqlalchemy'))


cpdef str to_str(s, encoding='utf-8'):
    if type(s) is str:
        return <str>s
    elif isinstance(s, bytes):
        return (<bytes>s).decode(encoding)
    elif isinstance(s, str):
        return str(s)
    elif s is None:
        return s
    else:
        raise TypeError(f"Could not convert from {s} to str.")


cpdef bytes to_binary(s, encoding='utf-8'):
    if type(s) is bytes:
        return <bytes>s
    elif isinstance(s, unicode):
        return (<unicode>s).encode(encoding)
    elif isinstance(s, bytes):
        return bytes(s)
    elif s is None:
        return None
    else:
        raise TypeError(f"Could not convert from {s} to bytes.")


cpdef unicode to_text(s, encoding='utf-8'):
    if type(s) is unicode:
        return <unicode>s
    elif isinstance(s, bytes):
        return (<bytes>s).decode('utf-8')
    elif isinstance(s, unicode):
        return unicode(s)
    elif s is None:
        return None
    else:
        raise TypeError(f"Could not convert from {s} to unicode.")


_type_dispatchers = WeakSet()


cdef class TypeDispatcher:
    def __init__(self):
        self._handlers = dict()
        self._lazy_handlers = dict()
        # store inherited handlers to facilitate unregistering
        self._inherit_handlers = dict()

        _type_dispatchers.add(self)

    cpdef void register(self, object type_, object handler):
        if isinstance(type_, str):
            self._lazy_handlers[type_] = handler
        elif isinstance(type_, tuple):
            for t in type_:
                self.register(t, handler)
        else:
            self._handlers[type_] = handler

    cpdef void unregister(self, object type_):
        if isinstance(type_, tuple):
            for t in type_:
                self.unregister(t)
        else:
            self._lazy_handlers.pop(type_, None)
            self._handlers.pop(type_, None)
            self._inherit_handlers.clear()

    cdef _reload_lazy_handlers(self):
        for k, v in self._lazy_handlers.items():
            mod_name, obj_name = k.rsplit('.', 1)
            mod = importlib.import_module(mod_name, __name__)
            self.register(getattr(mod, obj_name), v)
        self._lazy_handlers = dict()

    cpdef get_handler(self, object type_):
        cdef object clz, handler
        cdef dict d
        try:
            return self._handlers[type_]
        except KeyError:
            pass

        try:
            return self._inherit_handlers[type_]
        except KeyError:
            self._reload_lazy_handlers()
            for clz in type_.__mro__:
                if clz in self._handlers:
                    d = self._handlers
                elif clz in self._inherit_handlers:
                    d = self._inherit_handlers

                if clz in self._handlers:
                    handler = self._inherit_handlers[type_] = d[clz]
                    return handler
            raise KeyError(f'Cannot dispatch type {type_}')

    def __call__(self, object obj, *args, **kwargs):
        return self.get_handler(type(obj))(obj, *args, **kwargs)

    @staticmethod
    def reload_all_lazy_handlers():
        for dispatcher in _type_dispatchers:
            (<TypeDispatcher>dispatcher)._reload_lazy_handlers()


cdef inline build_canonical_bytes(tuple args, kwargs):
    if kwargs:
        args = args + (kwargs,)
    return pickle.dumps(tokenize_handler(args))


def tokenize(*args, **kwargs):
    return mmh_hash_bytes(build_canonical_bytes(args, kwargs)).hex()


def tokenize_int(*args, **kwargs):
    return mmh_hash(build_canonical_bytes(args, kwargs))


cdef class Tokenizer(TypeDispatcher):
    def __call__(self, object obj, *args, **kwargs):
        try:
            return self.get_handler(type(obj))(obj, *args, **kwargs)
        except KeyError:
            if hasattr(obj, '__mars_tokenize__') and not isinstance(obj, type):
                if len(args) == 0 and len(kwargs) == 0:
                    return obj.__mars_tokenize__()
                else:
                    obj = obj.__mars_tokenize__()
                    return self.get_handler(type(obj))(obj, *args, **kwargs)
            if callable(obj):
                if PDTick is not None and not isinstance(obj, PDTick):
                    return tokenize_function(obj)

            try:
                return cloudpickle.dumps(obj)
            except:
                raise TypeError(f'Cannot generate token for {obj}, type: {type(obj)}') from None


cdef inline list iterative_tokenize(object ob):
    cdef list dq = [ob]
    cdef int dq_pos = 0
    cdef list h_list = []
    while dq_pos < len(dq):
        x = dq[dq_pos]
        dq_pos += 1
        if type(x) in _primitive_types:
            h_list.append(x)
        elif isinstance(x, (list, tuple)):
            dq.extend(x)
        elif isinstance(x, set):
            dq.extend(sorted(x))
        elif isinstance(x, dict):
            dq.extend(sorted(x.items()))
        else:
            h_list.append(tokenize_handler(x))

        if dq_pos >= 64 and len(dq) < dq_pos * 2:  # pragma: no cover
            dq = dq[dq_pos:]
            dq_pos = 0
    return h_list


cdef inline tuple tokenize_numpy(ob):
    cdef int offset

    if not ob.shape:
        return str(ob), ob.dtype
    if hasattr(ob, 'mode') and getattr(ob, 'filename', None):
        if hasattr(ob.base, 'ctypes'):
            offset = (ob.ctypes.get_as_parameter().value -
                      ob.base.ctypes.get_as_parameter().value)
        else:
            offset = 0  # root memmap's have mmap object as base
        return (ob.filename, os.path.getmtime(ob.filename), ob.dtype,
                ob.shape, ob.strides, offset)
    if ob.dtype.hasobject:
        try:
            data = mmh_hash_bytes('-'.join(ob.flat).encode('utf-8', errors='surrogatepass'))
        except UnicodeDecodeError:
            data = mmh_hash_bytes(b'-'.join([to_binary(x) for x in ob.flat]))
        except TypeError:
            try:
                data = mmh_hash_bytes(pickle.dumps(ob, pickle.HIGHEST_PROTOCOL))
            except:
                # nothing can do, generate uuid
                data = uuid.uuid4().hex
    else:
        try:
            data = mmh_hash_bytes(ob.ravel().view('i1').data)
        except (BufferError, AttributeError, ValueError):
            data = mmh_hash_bytes(ob.copy().ravel().view('i1').data)
    return data, ob.dtype, ob.shape, ob.strides


cdef inline _extract_range_index_attr(object range_index, str attr):
    try:
        return getattr(range_index, attr)
    except AttributeError:  # pragma: no cover
        return getattr(range_index, '_' + attr)


cdef list tokenize_pandas_index(ob):
    cdef long long start
    cdef long long stop
    cdef long long end
    if isinstance(ob, pd.RangeIndex):
        start = _extract_range_index_attr(ob, 'start')
        stop = _extract_range_index_attr(ob, 'stop')
        step = _extract_range_index_attr(ob, 'step')
        # for range index, there is no need to get the values
        return iterative_tokenize([ob.name, getattr(ob, 'names', None), slice(start, stop, step)])
    else:
        return iterative_tokenize([ob.name, getattr(ob, 'names', None), ob.values])


cdef list tokenize_pandas_series(ob):
    return iterative_tokenize([ob.name, ob.dtype, ob.values, ob.index])


cdef list tokenize_pandas_dataframe(ob):
    l = [block.values for block in ob._data.blocks]
    l.extend([ob.columns, ob.index])
    return iterative_tokenize(l)


cdef list tokenize_pandas_categorical(ob):
    l = ob.to_list()
    l.append(ob.shape)
    return iterative_tokenize(l)


cdef list tokenize_pd_extension_dtype(ob):
    return iterative_tokenize([ob.name])


cdef list tokenize_categories_dtype(ob):
    return iterative_tokenize([ob.categories, ob.ordered])


cdef list tokenize_interval_dtype(ob):
    return iterative_tokenize([type(ob).__name__, ob.subtype])


cdef list tokenize_pandas_time_arrays(ob):
    return iterative_tokenize([ob.asi8, ob.dtype])


cdef list tokenize_pandas_tick(ob):
    return iterative_tokenize([ob.freqstr])


cdef list tokenize_pandas_interval_arrays(ob):
    return iterative_tokenize([ob.left, ob.right, ob.closed])


cdef list tokenize_sqlalchemy_data_type(ob):
    return iterative_tokenize([repr(ob)])


cdef list tokenize_sqlalchemy_selectable(ob):
    return iterative_tokenize([str(ob)])


cdef list tokenize_enum(ob):
    cls = type(ob)
    return iterative_tokenize([id(cls), cls.__name__, ob.name])


@lru_cache(500)
def tokenize_function(ob):
    if isinstance(ob, partial):
        args = iterative_tokenize(ob.args)
        keywords = iterative_tokenize(ob.keywords.items()) if ob.keywords else None
        return tokenize_function(ob.func), args, keywords
    else:
        try:
            if isinstance(ob, types.FunctionType):
                return iterative_tokenize([pickle.dumps(ob, protocol=0), id(ob)])
            else:
                return pickle.dumps(ob, protocol=0)
        except:
            pass
        try:
            return cloudpickle.dumps(ob, protocol=0)
        except:
            return str(ob)


@lru_cache(500)
def tokenize_pickled_with_cache(ob):
    return pickle.dumps(ob)


def tokenize_cupy(ob):
    from .serialization import serialize
    header, _buffers = serialize(ob)
    return iterative_tokenize([header, ob.data.ptr])


def tokenize_cudf(ob):
    from .serialization import serialize
    header, buffers = serialize(ob)
    return iterative_tokenize([header] + [(buf.ptr, buf.size) for buf in buffers])


cdef Tokenizer tokenize_handler = Tokenizer()

cdef set _primitive_types = {
    int, float, str, unicode, bytes, complex, type(None), type, slice, date, datetime, timedelta
}
for t in _primitive_types:
    tokenize_handler.register(t, lambda ob: ob)

for t in (np.dtype, np.generic):
    tokenize_handler.register(t, lambda ob: ob)

for t in (list, tuple, dict, set):
    tokenize_handler.register(t, iterative_tokenize)

tokenize_handler.register(np.ndarray, tokenize_numpy)
tokenize_handler.register(np.random.RandomState, lambda ob: iterative_tokenize(ob.get_state()))
tokenize_handler.register(memoryview, lambda ob: mmh3_hash_from_buffer(ob))
tokenize_handler.register(Enum, tokenize_enum)
tokenize_handler.register(pd.Index, tokenize_pandas_index)
tokenize_handler.register(pd.Series, tokenize_pandas_series)
tokenize_handler.register(pd.DataFrame, tokenize_pandas_dataframe)
tokenize_handler.register(pd.Categorical, tokenize_pandas_categorical)
tokenize_handler.register(pd.CategoricalDtype, tokenize_categories_dtype)
tokenize_handler.register(pd.IntervalDtype, tokenize_interval_dtype)
tokenize_handler.register(tzinfo, tokenize_pickled_with_cache)
tokenize_handler.register(pd.arrays.DatetimeArray, tokenize_pandas_time_arrays)
tokenize_handler.register(pd.arrays.TimedeltaArray, tokenize_pandas_time_arrays)
tokenize_handler.register(pd.arrays.PeriodArray, tokenize_pandas_time_arrays)
tokenize_handler.register(pd.arrays.IntervalArray, tokenize_pandas_interval_arrays)
tokenize_handler.register(pd.api.extensions.ExtensionDtype, tokenize_pd_extension_dtype)
if _has_cupy:
    tokenize_handler.register('cupy.ndarray', tokenize_cupy)
if _has_cudf:
    tokenize_handler.register('cudf.DataFrame', tokenize_cudf)
    tokenize_handler.register('cudf.Series', tokenize_cudf)
    tokenize_handler.register('cudf.Index', tokenize_cudf)

if PDTick is not None:
    tokenize_handler.register(PDTick, tokenize_pandas_tick)
if _has_sqlalchemy:
    tokenize_handler.register(
        "sqlalchemy.sql.sqltypes.TypeEngine", tokenize_sqlalchemy_data_type
    )
    tokenize_handler.register(
        "sqlalchemy.sql.Selectable", tokenize_sqlalchemy_selectable
    )

cpdef register_tokenizer(cls, handler):
    tokenize_handler.register(cls, handler)


@cython.nonecheck(False)
@cython.cdivision(True)
cpdef long long ceildiv(long long x, long long y) nogil:
    return x // y + (x % y != 0)


cdef class Timer:
    cdef object _start
    cdef readonly object duration

    def __enter__(self):
        self._start = time.time()
        return self

    def __exit__(self, *_):
        self.duration = time.time() - self._start


cdef mt19937_64 _rnd_gen
cdef bint _rnd_is_seed_set = False


cpdef void reset_id_random_seed() except *:
    cdef bytes seed_bytes
    global _rnd_is_seed_set

    seed_bytes = getrandbits(64).to_bytes(8, "little")
    _rnd_gen.seed((<uint_fast64_t *><char *>seed_bytes)[0])
    _rnd_is_seed_set = True


cpdef bytes new_random_id(int byte_len):
    cdef uint_fast64_t *res_ptr
    cdef uint_fast64_t res_data[4]
    cdef int i, qw_num = byte_len >> 3
    cdef bytes res

    if not _rnd_is_seed_set:
        reset_id_random_seed()

    if (qw_num << 3) < byte_len:
        qw_num += 1

    if qw_num <= 4:
        # use stack memory to accelerate
        res_ptr = res_data
    else:
        res_ptr = <uint_fast64_t *>malloc(qw_num << 3)

    try:
        for i in range(qw_num):
            res_ptr[i] = _rnd_gen()
        return <bytes>((<char *>&(res_ptr[0]))[:byte_len])
    finally:
        # free memory if allocated by malloc
        if res_ptr != res_data:
            free(res_ptr)


__all__ = ['to_str', 'to_binary', 'to_text', 'TypeDispatcher', 'tokenize', 'tokenize_int',
           'register_tokenizer', 'ceildiv', 'Timer', 'reset_id_random_seed', 'new_random_id']



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from typing import Any, Dict, List, Tuple
from ..lib import sparse
from ..oscar.debug import debug_async_timeout
from ..utils import (
    lazy_import,
    implements,
    register_ray_serializer,
)
from ..metrics import Metrics, Percentile, record_time_cost_percentile
from .base import StorageBackend, StorageLevel, ObjectInfo, register_storage_backend
from .core import BufferWrappedFileObject, StorageFileObject

ray = lazy_import("ray")


# TODO(fyrestone): make the SparseMatrix pickleable.


def _mars_sparse_matrix_serializer(value):
    return [value.shape, value.spmatrix]


def _mars_sparse_matrix_deserializer(obj) -> sparse.SparseNDArray:
    shape, spmatrix = obj
    return sparse.matrix.SparseMatrix(spmatrix, shape=shape)


def _register_sparse_matrix_serializer():
    # register a custom serializer for Mars SparseMatrix
    register_ray_serializer(
        sparse.matrix.SparseMatrix,
        serializer=_mars_sparse_matrix_serializer,
        deserializer=_mars_sparse_matrix_deserializer,
    )


class RayFileLikeObject:
    def __init__(self):
        self._buffers = []
        self._size = 0

    def write(self, content: bytes):
        self._buffers.append(content)
        self._size += len(content)

    def readinto(self, buffer):
        read_bytes = 0
        for b in self._buffers:
            read_pos = read_bytes + len(b)
            buffer[read_bytes:read_pos] = b
            read_bytes = read_pos
        return read_bytes

    def close(self):
        self._buffers.clear()
        self._size = 0

    def tell(self):
        return self._size


class RayFileObject(BufferWrappedFileObject):
    def __init__(self, object_id: Any, mode: str):
        super().__init__(object_id, mode, size=0)

    def _write_init(self):
        self._buffer = RayFileLikeObject()

    def _read_init(self):
        self._buffer = ray.get(self._object_id)
        self._mv = memoryview(self._buffer)
        self._size = len(self._buffer)

    def write(self, content: bytes):
        if not self._initialized:
            self._write_init()
            self._initialized = True

        return self._buffer.write(content)

    def _write_close(self):
        worker = ray.worker.global_worker
        metadata = ray.ray_constants.OBJECT_METADATA_TYPE_RAW
        args = [metadata, self._buffer.tell(), self._buffer, self._object_id]
        try:
            worker.core_worker.put_file_like_object(*args)
        except TypeError:
            args.append(None)  # owner_address for ray >= 1.3.0
            worker.core_worker.put_file_like_object(*args)

    def _read_close(self):
        pass


_support_specify_owner = None


def support_specify_owner():
    global _support_specify_owner
    if _support_specify_owner is None:
        sig = inspect.signature(ray.put)
        _support_specify_owner = "_owner" in sig.parameters
    return _support_specify_owner


@register_storage_backend
class RayStorage(StorageBackend):
    name = "ray"

    _storage_get_metrics = [
        (
            Percentile.PercentileType.P99,
            Metrics.gauge(
                "mars.storage.ray.get_cost_time_p99_seconds",
                "P99 time consuming in seconds to get object, every 1000 times report once.",
            ).record,
            1000,
        ),
        (
            Percentile.PercentileType.P95,
            Metrics.gauge(
                "mars.storage.ray.get_cost_time_p95_seconds",
                "P95 time consuming in seconds to get object, every 1000 times report once.",
            ).record,
            1000,
        ),
        (
            Percentile.PercentileType.P90,
            Metrics.gauge(
                "mars.storage.ray.get_cost_time_p90_seconds",
                "P90 time consuming in seconds to get object, every 1000 times report once.",
            ).record,
            1000,
        ),
    ]

    _storage_put_metrics = [
        (
            Percentile.PercentileType.P99,
            Metrics.gauge(
                "mars.storage.ray.put_cost_time_p99_seconds",
                "P99 time consuming in seconds to put object, every 1000 times report once.",
            ).record,
            1000,
        ),
        (
            Percentile.PercentileType.P95,
            Metrics.gauge(
                "mars.storage.ray.put_cost_time_p95_seconds",
                "P95 time consuming in seconds to put object, every 1000 times report once.",
            ).record,
            1000,
        ),
        (
            Percentile.PercentileType.P90,
            Metrics.gauge(
                "mars.storage.ray.put_cost_time_p90_seconds",
                "P90 time consuming in seconds to put object, every 1000 times report once.",
            ).record,
            1000,
        ),
    ]

    def __init__(self, *args, **kwargs):
        self._owner_address = kwargs.get("owner")
        self._owner = None  # A ray actor which will own the objects put by workers.

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        _register_sparse_matrix_serializer()
        return kwargs, dict()

    @staticmethod
    @implements(StorageBackend.teardown)
    async def teardown(**kwargs):
        pass

    @property
    @implements(StorageBackend.level)
    def level(self) -> StorageLevel:
        # TODO(fyrestone): return StorageLevel.MEMORY & StorageLevel.DISK
        # if object spilling is available.
        return StorageLevel.MEMORY | StorageLevel.REMOTE

    @implements(StorageBackend.get)
    async def get(self, object_id, **kwargs) -> object:
        if kwargs:  # pragma: no cover
            raise NotImplementedError(f'Got unsupported args: {",".join(kwargs)}')
        with debug_async_timeout(
            "ray_object_retrieval_timeout",
            "Storage get object timeout, ObjectRef: %s",
            object_id,
        ):
            with record_time_cost_percentile(self._storage_get_metrics):
                return await object_id

    @implements(StorageBackend.put)
    async def put(self, obj, importance=0) -> ObjectInfo:
        with record_time_cost_percentile(self._storage_put_metrics):
            if support_specify_owner() and self._owner_address:
                if not self._owner:
                    self._owner = ray.get_actor(self._owner_address)
                object_id = ray.put(obj, _owner=self._owner)
            else:
                object_id = ray.put(obj)
        # We can't get the serialized bytes length from ray.put
        return ObjectInfo(object_id=object_id)

    @implements(StorageBackend.delete)
    async def delete(self, object_id):
        ray.internal.free(object_id)

    @implements(StorageBackend.object_info)
    async def object_info(self, object_id) -> ObjectInfo:
        # The performance of obtaining the object size is poor.
        return ObjectInfo(object_id=object_id)

    @implements(StorageBackend.open_writer)
    async def open_writer(self, size=None) -> StorageFileObject:
        new_id = ray.ObjectRef.from_random()
        ray_writer = RayFileObject(new_id, mode="w")
        return StorageFileObject(ray_writer, object_id=new_id)

    @implements(StorageBackend.open_reader)
    async def open_reader(self, object_id) -> StorageFileObject:
        ray_reader = RayFileObject(object_id, mode="r")
        return StorageFileObject(ray_reader, object_id=object_id)

    @implements(StorageBackend.list)
    async def list(self) -> List:
        raise NotImplementedError("Ray storage does not support list")

    @implements(StorageBackend.fetch)
    async def fetch(self, object_id):
        pass
